/*
 *      $Header: adc.c,v 0.5 1995/10/24 16:32:20 amish$
 *      Revision 0.5  1996/10/9 11:33:00  amish
 *      Revision 0.5
 *
 *      Revision 0.6  1998/4/6 17:30:00  amish
 *      Revision 0.6
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/dma.h>

#include "adc.h"
#include "atMIO16x.h"

#define MAX_DMA_SIZE	(128 * 1024)

#ifndef ADCBUFSIZE
#define ADCBUFSIZE	32768
#endif

static struct semaphore mutex = MUTEX;

int adcbufsize = ADCBUFSIZE;

/* Accessing Board command and status registers */
#define ADCstatus( board )		\
	(readport_w( &((board)->regs.stat1) ))

/* Clear the ADC - strobe clear register */
#define ClearADC( adc )			\
	inb_p( DAQ_DAQCLR )

/* Start an ADC conversion by reading from the ADC start register */
#define StartADCConversion( adc )	\
	inb_p( DAQ_DAQSTRT )

/* Read a 16-bit ADC conversion value from the ADC FIFO */
#define ReadADCFIFO( adc )		\
	inw( DAQ_FIFO )

/* Clear an ADC DMA Terminal Count interrupt */
#define ClearDMAInterrupt( adc )	\
	outb_p( 0x00, (adc)->dmaintclr )

/* Disable all ADC interrupt generation */
#define DisableADCInterrupts( adc )		\
	clearbits_w( &((board)->regs.cmd3), DAQCMPLINT | IO_INT | ADCREQ | DMACHA )

#define Reset9513Counter(ctr) \
{\
	outw_p(0xffc0 + (1<<((ctr)-1)), DAQ_9513CMD);	\
	outw_p(0xff00 + (ctr), DAQ_9513CMD);	\
	outw_p(0x0004, DAQ_9513DATA);	\
	outw_p(0xff08 + (ctr), DAQ_9513CMD);	\
	outw_p(0x0003, DAQ_9513DATA);	\
	outw_p(0xff40 + (1<<((ctr)-1)), DAQ_9513CMD);	\
	outw_p(0xff40 + (1<<((ctr)-1)), DAQ_9513CMD);	\
}

#define STATIC

static int eeprom_shift_out_mode(int addr);
static int eeprom_shift_in_val(void);
static void write_caldac(int dac, int value);

STATIC int allocate_adc_buffers( ADC_t *adc )
{
	if (adc->bufsize == 0)
	{
		adc->frontbuf = NULL;
		adc->backbuf = NULL;
		return 0;
	}
	printk( "atMIO16x: Using DMA buffer size, %d\n", adc->bufsize );

	/* Allocate buffers for DMA transfers */
	if (adc->bufsize * sizeof( short ) > MAX_DMA_SIZE)
	{
		printk( "atMIO16x: Requested DMA buffer size, %d, too large\n", adc->bufsize );
		adc->bufsize = MAX_DMA_SIZE / sizeof( short );
		printk( "atMIO16x: Using DMA buffer size, %d\n", adc->bufsize );
	}
	adc->frontbuf = kmalloc( adc->bufsize * sizeof( short ), GFP_KERNEL | GFP_DMA );
	if (adc->frontbuf == NULL)
	{
		printk( "ADC: Unable to allocate DMA buffer (%d)\n", adc->bufsize * sizeof(short) );
		adc->bufsize = 0;
		return -(ENOMEM);
	}
	adc->backbuf = kmalloc( adc->bufsize * sizeof( short ), GFP_KERNEL | GFP_DMA );
	if (adc->backbuf == NULL)
	{
		kfree( adc->frontbuf );
		adc->bufsize = 0;
		printk( "ADC: Unable to allocate DMA buffer (%d)\n", adc->bufsize * sizeof(short) );
		return -(ENOMEM);
	}
	return 0;
}

STATIC int read_adc( struct inode *inode, struct file *file, char *buffer, int count )
{
	ADC_t *adc;
	int retval = 0;
	unsigned long flags;

	/* Which ADC? */
	if ((adc = GETADC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	count /= sizeof( short );
/*	printk("count = %d, bufsize = %d, incr = %d\n", count, adc->bufsize, adc->incr); */
	if (count < adc->bufsize)
		return -(EINVAL);

	/* Lockout other processes from accessing the ADC */
	down( &(adc->sem) );

	/*
	** (indx != 0) tells us we have a complete buffer (in backbuf)
	** that can be transferred to the user. (started != 0) tells
	** us we have either not started, or, if (incr != 0), that there
	** is a partial buffer (because acquisition was stopped before
	** completing a buffer)
	*/

	/*
	** If we are started, but we don't have a complete buffer of
	** data, then we wait (unless O_NONBLOCK)
	*/
	save_flags( flags );
	cli();
	retval = 0;
	if ((adc->started != 0) && (adc->indx == 0))
	{
		if (file->f_flags & O_NONBLOCK)
		{
			retval = -(EAGAIN);
		}
		else
		{
			interruptible_sleep_on( &(adc->wait) );
/*			printk("returned from sleep\n"); */
			if (current->signal & ~current->blocked)
			{
				ClearADC( adc );
				dprintk(( "ADC: interrupted\n" ));
				retval = -(ERESTARTSYS);
			}
		}
	}
	if (retval == 0)
	{
		if (adc->overrun)
		{
			retval = -(EIO);
			printk( "ADC: overrun detected\n" );
		}
		else if (adc->overflow)
		{
			retval = -(EIO);
			printk( "ADC: overflow detected\n" );
		}
		else if (adc->indx)
		{
			restore_flags( flags );
			/* A buffer has filled - copy it to user space */
			memcpy_tofs( buffer, adc->backbuf, adc->indx * sizeof( short ) );
			buffer += adc->indx * sizeof( short );
/*			dprintk(( "ADC: %d words transferred\n", adc->indx )); */
			retval = adc->indx * sizeof( short );
			adc->indx = 0;
			/* Release the exclusion semaphore */
			up( &(adc->sem) );
			return (retval);
		}
		else if ((adc->started == 0) && (adc->incr))
		{
			restore_flags( flags );
			memcpy_tofs( buffer, adc->frontbuf, adc->incr * sizeof( short ) );
			buffer += adc->incr * sizeof( short );
/*			dprintk(( "ADC: %d words transferred\n", adc->incr )); */
			retval = adc->incr * sizeof( short );
			adc->incr = 0;
			/* Release the exclusion semaphore */
			up( &(adc->sem) );
			return (retval);
		}
		else if ((adc->started == 0) && (adc->incr == 0))
		{
			retval = -(ENODATA);
		}
		else
		{
			printk( "atMIO16x: DMA buffer overflow\n" );
			retval = -(EIO);
		}
	}

	restore_flags( flags );
	/* Release the exclusion semaphore */
	up( &(adc->sem) );
	return (retval);
}

STATIC int select_adc(struct inode *inode, struct file *file, int sel_type, select_table *wait)
{
	ADC_t *adc = BOARD(MAJOR(inode->i_rdev))->adc;

	/*
	** We are always ready for output or exceptions (ie: never wait on output
	** or exceptions.  If there is a buffer of data to be transferred, then we
	** don't wait on input either.  In all other cases, we wait.
	** This means we call select_wait() (which returns immediately), and return 0,
	** so that our calling function (do_select()) can wait.
	*/
	if (sel_type != SEL_IN)
		return 1;
	if ((adc->indx != 0) || ((adc->started == 0) && (adc->incr != 0)))
		return 1;
	cli();
	select_wait( &(adc->wait), wait );
	sti();
	return 0;
}

STATIC int init_adc( ADC_t *adc )
{
	/* Initialise the ADC registers */
	DisableADCInterrupts(adc);
	ClearADC(adc);

	/* Set the data ports for the ADC */
	adc->dmaintclr = DAQ_DMATCACLR;

	/* Initialise options and state variables */
	adc->nchans	= 0;
	adc->started = 0;
	adc->total = 0;
	adc->samplerate	= 20000;
	adc->clockspeed	= 1000000;
	adc->watchdog = 0;
	adc->sampcnt = 2;
	adc->triggermode = ADC_TRIGGERMODE_POST;
	adc->usedma = 1;

	/* Access control variables */
	adc->sem = mutex;
	adc->wait = NULL;
	adc->overflow = 0;
	adc->overrun = 0;

	return 0;
}

STATIC int start_adc( ADC_t *adc )
{
	int i, interval, sampcnt;
	int clockspeed = adc->clockspeed;

	dprintk(( "atMIO16x: start_adc() called\n" ));
	dprintregs();

	adc->overrun = 0;
	adc->overflow = 0;

	/*
	** Before doing anything else,
	** validate the desired # of chans and samplerate
	** with the actual capability of the board
	*/
	dprintk(( "atMIO16x: 1start_adc(): rate=%d, nchans=%d, clock=%d\n", adc->samplerate, adc->nchans, adc->clockspeed ));
	if (adc->nchans == 0)
		return -(EINVAL);
	if (adc->samplerate * adc->nchans > 100000)
		return -(EINVAL);
	interval = clockspeed / (adc->samplerate * adc->nchans);
	if (interval * adc->samplerate * adc->nchans != clockspeed)
		return -(EINVAL);
	dprintk(( "atMIO16x: start_adc(): rate = %d, nchans = %d, interval = %d\n", adc->samplerate, adc->nchans, interval ));

	/*
	** Reset the Am9513A counters
	*/
	Reset9513Counter(3);
	Reset9513Counter(4);
/*	Reset9513Counter(5); */

	/*
	** Load the desired channel scan sequence into the
	** channel configuration memory
	*/
	inb_p( DAQ_CONFIGMEMCLR );
	adc->cam[ adc->nchans - 1] |= CHAN_LAST;
	for (i=0; i < adc->nchans; i++)
		outw_p( adc->cam[i], DAQ_CFGMEM );
	outb_p( 0x00, DAQ_CONFIGMEMLD );

	/*
	** Program the sample-interval counter (Cntr3 on the Am9513A)
	*/
	outw_p( 0xff03, DAQ_9513CMD );		/* Select cntr3 mode reg. */

	if (clockspeed == 5000000)
		outw_p( 0x8225, DAQ_9513DATA );	/* Select 5Mhz clock */
	else if (clockspeed == 1000000)
		outw_p( 0x8b25, DAQ_9513DATA );	/* Select 1Mhz clock */
	else if (clockspeed == 100000)
		outw_p( 0x8c25, DAQ_9513DATA );	/* Select 100kHz clock */
	else if (clockspeed == 10000)
		outw_p( 0x8d25, DAQ_9513DATA );	/* Select 10kHz clock */
	else if (clockspeed == 1000)
		outw_p( 0x8e25, DAQ_9513DATA );	/* Select 1kHz clock */
	else if (clockspeed == 100)
		outw_p( 0x8f25, DAQ_9513DATA );	/* Select 100Hz clock */
	else
	{
		dprintk(( "atMIO16x: INVALID CLOCK SELECTION-using 100kHz!!\n" ));
		outw_p( 0x8c25, DAQ_9513DATA );	/* Select 100kHz clock */
	}

	outw_p( 0xff0b, DAQ_9513CMD );		/* Select cntr3 load reg. */
	outw_p( 0x0002, DAQ_9513DATA );		/* Store cntr3 load value */
	outw_p( 0xff44, DAQ_9513CMD );		/* Load cntr3 */
	outw_p( 0xfff3, DAQ_9513CMD );		/* Step cntr3 down to 1 */
	outw_p( interval, DAQ_9513DATA );	/* Write desired interval */
	outw_p( 0xff24, DAQ_9513CMD );		/* Arm Counter 3 */

	/*
	** Program the sample counter: (Counter4 on the Am9513A)
	** this involves setting it up for pretriggering mode,
	** which means that it will collect indefinitely
	** (w/o counting) until a second trigger is received,
	** whereupon it will count down the following value
	** before stopping.
	*/
	if (adc->sampcnt <= 65536)
	{
		clearbits_w( &((board)->regs.cmd1), CNT32_16 );	/* Inform DAQ that cntr5 is free */
		outw_p( 0xff04, DAQ_9513CMD );	/* Select cntr4 mode reg. */
		if (adc->triggermode == ADC_TRIGGERMODE_PRE)
			outw_p( 0x9025, DAQ_9513DATA );	/* use Pretrigger mode */
		else
			outw_p( 0x1025, DAQ_9513DATA );	/* use Posttrigger mode */
		outw_p( 0xff0c, DAQ_9513CMD );	/* Select cntr4 load reg. */
		sampcnt = adc->sampcnt;
		outw_p( sampcnt, DAQ_9513DATA );	/* set sample count */
		outw_p( 0xff48, DAQ_9513CMD );	/* Load cntr4 */
		outw_p( 0xfff4, DAQ_9513CMD );	/* decrement cntr4 */
		outw_p( 0xff28, DAQ_9513CMD );	/* Arm Counter 4 */
	}
	else
	{
#ifdef notdef
		/*
		** This code sets up the transfer of a fixed number of samples, where
		** that number is encoded by 32-bits.
		**
		** This is all ifdef'ed out, because it doesn't work.
		*/
		setbits_w( &((board)->regs.cmd1), CNT32_16 ); /* Enable 32 bit sample counter */
		outw_p( 0xff04, DAQ_9513CMD );	/* Select cntr4 mode reg. */
		if (adc->triggermode == ADC_TRIGGERMODE_PRE)
			outw_p( 0x9025, DAQ_9513DATA );	/* use Pretrigger mode */
		else
			outw_p( 0x1025, DAQ_9513DATA );	/* use Posttrigger mode */
		outw_p( 0xff0c, DAQ_9513CMD );	/* Select cntr4 load reg. */
		sampcnt = (adc->sampcnt & 0x0000ffff) - 1;
		printk("sampcnt(LSB) = 0x%x\n", sampcnt);
		outw_p( sampcnt, DAQ_9513DATA );	/* set sample count (16LSB) */
		outw_p( 0xff48, DAQ_9513CMD );	/* Load cntr4 */
		outw_p( 0, DAQ_9513DATA );		/* Store 0 into cntr4 Load reg. */
		outw_p( 0xff28, DAQ_9513CMD );	/* Select cntr5 mode reg. */
		outw_p( 0xff05, DAQ_9513CMD );	/* Select cntr5 mode reg. */
		outw_p( 0x0025, DAQ_9513DATA );	/* Store cntr5 mode value */
		outw_p( 0xff0d, DAQ_9513CMD );	/* Select cntr5 Load reg. */
		sampcnt = adc->sampcnt & 0x0000ffff;
		if ((sampcnt == 0) || (sampcnt == 1))
			sampcnt = 0;
		else sampcnt = 1;
			sampcnt += (adc->sampcnt & 0xffff0000) >> 16;
		printk("sampcnt(MSB) = 0x%x\n", sampcnt);
		outw_p( sampcnt, DAQ_9513DATA );	/* set sample count (16MSB) */
		outw_p( 0xff70, DAQ_9513CMD );	/* Load & Arm Counter 5 . */
		printk("ADC: 32-bit sample count not supported\n");
#endif
	}

#ifdef notdef
	/*
	** Generate interrupts only when FIFO is half full (rather than
	** whenever it is not completely empty)
	*/
	setbits_w( &((board)->regs.cmd4), ADCFIFOREQ );
#endif


	/*
	** Start acquisition (after trigger)
	*/
	setbits_w( &((board)->regs.cmd1), SCANEN | DAQEN ); /* Enab. acqu. & scanning */


	/*
	** Now, only the DMA business and the actual triggering are needed.
	** When setting up the ADC DMA stuff on Channel A, be careful not to
	** affect the DAC stuff that might be using Channel B.
	** (see 4-13, 4-14 in AT-MIO-16X User Manual)
	*/
	if (adc->usedma)
	{
		/* Enable DMA IRQs on chan A */
		ClearDMAInterrupt(adc);
		setbits_w(&((board)->regs.cmd3), DAQCMPLINT | DMATCINT | ADCREQ | DMACHA);

		/* Program the DMA controller */
		cli();
		disable_dma( board->dma );
		adc->count = adc->bufsize;
		adc->indx = adc->incr = 0;
		clear_dma_ff( board->dma );
		set_dma_mode( board->dma, DMA_MODE_READ );
		set_dma_addr( board->dma, (unsigned long)adc->frontbuf );
		set_dma_count( board->dma, adc->count * sizeof(short) );
		enable_dma( board->dma );
		sti();
	}
	else
	{
		adc->count = adc->bufsize;
		adc->indx = adc->incr = 0;
		setbits_w(&((board)->regs.cmd3), DAQCMPLINT | ADCREQ | IO_INT);
	}

	/* NOTE: Collection doesn't start now - it will start when the ADC is triggered. */
	adc->started = 1;
	adc->total = 0;

	dprintk(( "atMIO16x: start_adc() finished\n" ));
	dprintregs();
	return 0;
}

STATIC int stop_adc( ADC_t *adc )
{
	/*
	** Because we are in pretrigger mode, this will cause
	** the sample counter to start decrementing.  Since we
	** set that to 2 (the minimum), we will get two more
	** samples before acquisition stops
	*/
	StartADCConversion(adc);
	return 0;
}

/* NOTE: Can be called with interrupts disabled - don't do anything to re-enable them! */
int trigger_adc( ADC_t *adc )
{
	/*
	*/
	StartADCConversion(adc);
	return 0;
}

STATIC int ioctl_adc(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	ADC_t 	*adc;
	int	stat = 0, i;
	ADC_chan_t tmpchan;
	unsigned long addr, value;

	if ((stat = atmio16x_ioctl_check( cmd, arg )) != 0)
		return stat;

	/* Which ADC? */
	if ((adc = GETADC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	switch (cmd)
	{
		case ADC_CLEARCHANS:
			down( &(adc->sem) );
			adc->nchans = 0;
			dprintk(( "atMIO16x: handled CLEARCHANS ioctl\n" ));
			up( &(adc->sem) );
			break;
		case ADC_ADDCHAN:
			down( &(adc->sem) );
			i = adc->nchans;
			memcpy_fromfs(&tmpchan, (void *)arg, sizeof(ADC_chan_t));
			if ((i == ADC_MAXCHANS) || (tmpchan.channel < 16))
			{
				adc->cam[i] = 0;
				adc->cam[i] |= tmpchan.channel << 6;
				if (tmpchan.gain < 2) tmpchan.gain = 1;
				else if (tmpchan.gain < 5) tmpchan.gain = 2;
				else if (tmpchan.gain < 10) tmpchan.gain = 3;
				else if (tmpchan.gain < 20) tmpchan.gain = 4;
				else if (tmpchan.gain < 50) tmpchan.gain = 5;
				else if (tmpchan.gain < 100) tmpchan.gain = 6;
				else tmpchan.gain = 7;
				adc->cam[i] |= tmpchan.gain << 3;
				if (tmpchan.ghost) adc->cam[i] |= CHAN_GHOST;
				if (!tmpchan.unipolar) adc->cam[i] |= CHAN_BIP;
				if (tmpchan.inputmode != ADC_INPUTMODE_DIFF) adc->cam[i] |= CHAN_SE;
				if (tmpchan.inputmode == ADC_INPUTMODE_NRSE) adc->cam[i] |= CHAN_AIS;
				adc->nchans++;
				dprintk(( "atMIO16x: handled ADDCHAN ioctl (0x%x) (n = %d)\n", (0xffff & adc->cam[i]), adc->nchans ));
			}
			else stat = -(EINVAL);
			up( &(adc->sem) );
			break;
		case ADC_SETCNT:
			down( &(adc->sem) );
			adc->sampcnt = 1 + get_fs_long((unsigned long *)arg);
			up( &(adc->sem) );
			break;
		case ADC_SETCLOCK:
			down( &(adc->sem) );
			adc->clockspeed = get_fs_long((unsigned long *)arg);
			up( &(adc->sem) );
			break;
		case ADC_GETCLOCK:
			put_fs_long( adc->clockspeed, (unsigned long *)arg );
			break;
		case ADC_SETSR:
			down( &(adc->sem) );
			adc->samplerate = get_fs_long((unsigned long *)arg);
			if (adc->samplerate * adc->nchans > 100000)
				stat = -(EINVAL);
			up( &(adc->sem) );
			break;
		case ADC_GETSR:
			put_fs_long( adc->samplerate, (unsigned long *)arg );
			break;
		case ADC_STARTACQ:
			down( &(adc->sem) );
			start_adc( adc );
			dprintk(( "atMIO16x: handled STARTACQ ioctl\n" ));
			up( &(adc->sem) );
			break;
		case ADC_STOPACQ:
			down( &(adc->sem) );
			stop_adc( adc );
			dprintk(( "atMIO16x: handled STOPACQ ioctl\n" ));
			up( &(adc->sem) );
			break;
		case ADC_SETBUFSIZE:
			if (arg * sizeof( short ) > MAX_DMA_SIZE)
			{
				return -(EINVAL);
			}
			else
			{
				down( &(adc->sem) );
				if (adc->bufsize)
				{
					kfree( adc->frontbuf );
					kfree( adc->backbuf );
				}	
				adc->bufsize = get_fs_long((unsigned long *)arg);
				stat = allocate_adc_buffers( adc );
				up( &(adc->sem) );
			}
			break;
		case ADC_GETBUFSIZE:
			put_fs_long( adc->bufsize, (unsigned long *)arg );
			break;
		case ADC_READFIFO:
			i = inw_p(DAQ_FIFO);
			dprintk(( "atMIO16x: FIFO val = %d\n", i ));
			put_fs_long(i, (unsigned long *)arg );
			break;
		case ADC_USEDMA:
			adc->usedma = get_fs_long((unsigned long *)arg);
			printk("usedma = %d\n", adc->usedma);
			break;
		case ADC_SETTRIGMODE:
			adc->triggermode = get_fs_long((unsigned long *)arg);
			printk("triggermode = %d\n", adc->triggermode);
			break;
		case ADC_TRIGGER:
			trigger_adc(adc);
			dprintk(( "atMIO16x: triggered\n" ));
			dprintregs();
			break;
		case ADC_READEEPROM:
			dprintk(( "atMIO16x: called READEEPROM...\n" ));
			stat = -(EINVAL);
			addr = get_fs_long((unsigned long *)arg);
			if (eeprom_shift_out_mode(addr) != -1)
			{
				value = eeprom_shift_in_val();
				put_fs_long(value, (unsigned long *)arg);
				dprintk(( "atMIO16x: READEEPROM done. Found %lx at %ld.\n", value, addr ));
				stat = 0;
			}
			else dprintk(( "atMIO16x: ERROR reading EEPROM.\n" ));
			break;
		case ADC_CALIBRATE:
			dprintk(( "atMIO16x: initiating ADC calibration...\n" ));
			outb_p(0x00, DAQ_ADCCAL);
			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + (15 * (HZ/10));
			outb_p(0x00, DAQ_SINGCONV);
			ClearADC(adc);
			dprintk(( "atMIO16x: finished ADC calibration procedure.\n" ));
			break;
		case ADC_SCONV:
			dprintk(( "atMIO16x: generating SINGCONV\n" ));
			outb_p(0x00, DAQ_SINGCONV);
			{
				int jj=1;
				while ((!(ADCstatus(board) & ADCFIFOEF)) && (jj != 0))
					jj++;
				jj = inw_p(DAQ_FIFO);
				dprintk(( "atMIO16x: FIFO val = %d\n", jj ));
			}
			break;
		case ADC_INIT:
			DisableADCInterrupts( adc );
			ClearADC( adc );
			dprintk(( "atMIO16x: handled INIT\n" ));
			dprintregs();
			break;
		case ADC_STATUS:
			/* Read the ADC status register - returning byte read */
			dprintregs();
			dprintk(( "watchdog = %d, stat2 = 0x%x, %d %d\n", adc->watchdog, 1 & inw_p(DAQ_STAT2),
				((short *)(adc->frontbuf))[0], ((short *)(adc->frontbuf))[1] ));
			put_fs_long((unsigned long)(ADCstatus(board)), (unsigned long *)arg);
			break;
		default:
			return -(ENOSYS);
			break;
	}

	return stat;
}

STATIC int open_adc( struct inode *inode, struct file *file )
{
	ADC_t 	*adc;

	dprintk(( "atMIO16x: open_adc() called\n" ));

	/* Which ADC? */
	if ((adc = GETADC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	adc->open++;

	return 0;
}

STATIC void release_adc(struct inode *inode, struct file *file)
{
	ADC_t 	*adc;
	int		minor;

	dprintk(( "atMIO16x: release_adc() called\n" ));

	minor = MINOR(inode->i_rdev);
	if ((adc = GETADC( minor )) == NULL)
		return;

	adc->open--;
	if (adc->open == 0)
	{
		/* Last close - shutdown the ADC if in running mode */
		if (ADCstatus( board ) & (DAQPROG | OVERRUN | OVERFLOW))
			ClearADC( adc );
	}
	atmio16x_close( inode, file );
}

void adc_interrupt( ADC_t *adc, int status )
{
	short	*buf;

/*	printk("atMIO16x: ADC IRQ: 0x%x\n", status); */
	adc->watchdog++;
	if (status & (OVERRUN | OVERFLOW))
	{
		dprintregs();
		DisableADCInterrupts( adc );
		ClearADC( adc );
		if (status & OVERRUN)
		{
			adc->overrun = 1;
			printk( "atMIO16x: ADC overrun\n" );
		}
		if (status & OVERFLOW)
		{
			adc->overflow = 1;
			printk( "atMIO16x: ADC overflow\n" );
		}
		wake_up_interruptible( &(adc->wait) );
	}
	else if (status & DMATCA)
	{
		/* Check the back buffer has been emptied */ 
		if (adc->indx)
		{
			printk( "atMIO16x: ADC DMA buffer overflow\n" );
			adc->indx = 0;
			ClearDMAInterrupt( adc );
			wake_up_interruptible( &(adc->wait) );
			return;
		}
		buf = adc->frontbuf;
		adc->frontbuf = adc->backbuf;
		adc->backbuf = buf;
		adc->indx = adc->count;

		/* setup next DMA transfer */
		/*
		** Since I'm using 'FAST' interrupts, I don't need to cli()/sti() this
		** stuff.  If I wasn't, I'd put:
		**   cli(); disable_dma(board->dma);
		** here, and 
		**   clear_dma_ff(board->dma); set_dma_mode(board->dma, DMA_MODE_READ);
		** before set_dma_addr(), and
		**   sti()
		** after enable_dma(), below.
		*/
		if (get_dma_residue(board->dma) != 0)
			printk("atMIO16x: unexpected DMA residue = %d\n", get_dma_residue(board->dma));
		adc->count = adc->bufsize;
		set_dma_addr( board->dma, (unsigned long)adc->frontbuf );
		set_dma_count( board->dma, adc->count * sizeof( short ) );
		enable_dma( board->dma );

		ClearDMAInterrupt( adc );
		wake_up_interruptible( &(adc->wait) );
	}
	else if ((adc->usedma == 0) && ((status & ADCFIFOHF) == 0))
	{
		buf = adc->frontbuf;
		for (adc->cnt = 0; adc->cnt < 256; adc->cnt++)
		{
			buf[adc->incr] = ReadADCFIFO(adc);
			adc->incr++;
			adc->total++;
			if (adc->incr >= adc->count)
			{
				if (adc->indx)
				{
					printk("atMIO16x: ADC buffer overflow\n");
				}
				adc->frontbuf = adc->backbuf;
				adc->backbuf = buf;
				buf = adc->frontbuf;
				adc->indx = adc->count;
				adc->incr = 0;
				wake_up_interruptible( &(adc->wait) );
			}
		}
		/*		printk("atMIO16x: DAQ IRQ: %d\n", adc->incr); */
	}
	else if ((adc->usedma == 0) && ((status & ADCFIFOEF) == 0))
	{
		buf = adc->frontbuf;
		do {
			buf[adc->incr] = ReadADCFIFO(adc);
			adc->incr++;
			adc->total++;
			if (adc->incr >= adc->count)
			{
				if (adc->indx)
				{
					printk("atMIO16x: ADC buffer overflow\n");
				}
				adc->frontbuf = adc->backbuf;
				adc->backbuf = buf;
				buf = adc->frontbuf;
				adc->indx = adc->count;
				adc->incr = 0;
				wake_up_interruptible( &(adc->wait) );
			}
		} while (ADCstatus(board) & ADCFIFOEF);
		printk("atMIO16x: DAQ IRQ: %d\n", adc->total);
	}
	else if ((status & DAQCOMP) == 0)
	{
/*		dprintk(( "atMIO16x: ADC interrupt for no good reason\n" )); */
	}
	if (status & DAQCOMP)
	{
		if (adc->usedma)
		{
			disable_dma(board->dma);
			printk("atMIO16x: DAQCOMP received, residue = %d\n", get_dma_residue(board->dma));
			adc->incr = adc->count - (get_dma_residue(board->dma) / 2) - 1;
			ClearDMAInterrupt(adc);
		}
		dprintregs();
		DisableADCInterrupts(adc);

		/*
		** The code below is supposed to handle the situation of receiving a DAQCOMP
		** signal from the board while there is still data in the ADC buffer.  I
		** commented it out while debugging, and haven't gotten around to re-enabling
		** it (i.e.: it needs to be tested)
		*/
/*		buf = adc->frontbuf;
		while (ADCstatus(board) & ADCFIFOEF)
		{
			printk("atMIO16x: DAQCOMP: ADC buffer still had data in it\n");
			buf[adc->incr] = ReadADCFIFO(adc);
			adc->incr++;
			adc->total++;
			if (adc->incr >= adc->count)
			{
				if (adc->indx)
				{
					printk("atMIO16x: ADC buffer overflow\n");
				}
				adc->frontbuf = adc->backbuf;
				adc->backbuf = buf;
				buf = adc->frontbuf;
				adc->indx = adc->count;
				adc->incr = 0;
				wake_up_interruptible(&(adc->wait));
			}
		}
*/
		adc->started = 0;
		ClearADC(adc);
		wake_up_interruptible(&(adc->wait));
	} 
	return;
}

struct file_operations adc_fops =
{
	NULL,			/* seek */
	read_adc,
	NULL,			/* write */
	NULL,			/* readdir */
	select_adc,
	ioctl_adc,		/* ioctl */
	NULL,			/* mmap */
	open_adc,
	release_adc,
};

/* Functions to load and unload the ADC structure */

ADC_t *get_adc( int minor )
{
	ADC_t *adc;

	adc = (ADC_t *)kmalloc(sizeof(ADC_t), GFP_KERNEL);
	adc->open = 0;
	init_adc(adc);

	adc->bufsize = adcbufsize;
	if (allocate_adc_buffers(adc) < 0)
	{
		kfree(adc);
		return NULL;
	}
	adc->indx = adc->count = 0;

	return adc;
}

void free_adc( ADC_t *adc )
{
	if (adc != NULL)
	{
		if (adc->frontbuf != NULL)
		{
			kfree(adc->frontbuf);
			adc->frontbuf = NULL;
		}
		if (adc->backbuf != NULL)
		{
			kfree(adc->backbuf);
			adc->backbuf = NULL;
		}
		init_adc(adc);
		kfree(adc);
	}
}


/*
** Settings for reading the calibration constants from the EEPROM, and writing them to the
** appropriate calibrations DACs
*/
#define EEPROM_READ 0x180
#define MODE_CNT 9

#define AREA_INFORMATION          10
#define ADC_PREG_OFFSET_MSB       9
#define ADC_PREG_OFFSET_LSB       8
#define DAC1_OFFSET               7
#define DAC0_OFFSET               6
#define DAC1_GAIN                 5
#define DAC0_GAIN                 4
#define ADC_POSTG_OFFSET_FINE     3
#define ADC_POSTG_OFFSET_COURSE   2
#define ADC_GAIN_FINE             1
#define ADC_GAIN_COURSE           0

#define DAC_IN_GAIN_COURSE          0
#define DAC_IN_GAIN_FINE            1
#define DAC_IN_POSTG_OFFSET_COURSE  2
#define DAC_IN_POSTG_OFFSET_FINE    3
#define DAC_OUT0_GAIN               4
#define DAC_OUT1_GAIN               5
#define DAC_OUT0_OFFSET             6
#define DAC_OUT1_OFFSET             7
#define DAC_IN_PREG_OFFSET          8

#define SDATA_BIT_POS     14

static int eeprom_shift_out_mode(int addr)
{
	int prom_input, index, rc = 0;
	unsigned int bitmask, data;

	/*
	 * Pulse the SCLK bit in Command Register 1 high to initialize the EEPROM
	 * circuitry.
	 */
	writebits_w( &((board)->regs.cmd1), 0, SCLK | EEPROMCS | SDATA);					/* Clear SCLK and EEPROMCS */
	writebits_w( &((board)->regs.cmd1), SCLK, EEPROMCS);								/* Pulse SCLK ON */
	writebits_w( &((board)->regs.cmd1), EEPROMCS, SCLK | SDATA);						/* Pulse SCLK OFF, Set EEPROMCS - Enable chip select */
	writebits_w( &((board)->regs.cmd1), EEPROMCS | SCLK, SDATA);						/* Pulse SCLK ON */
	
	/*
	 * The EEPROM_READ mode data is clocked into the EEPROM one bit at a time
	 * with a low-to-high transition of the SCLK bit.  The serial data bits are
	 * set/cleared in the SDATA bit of the Command Register 1.
	 */
	data = addr | EEPROM_READ;
	bitmask = 0x0100;
	for ( index = 0; index < MODE_CNT; index++)
	{
		prom_input = ((data & (bitmask >> index)) ? 1 : 0);
		if (prom_input == 0)
		{
			writebits_w( &((board)->regs.cmd1), EEPROMCS, SCLK | SDATA);				/* Pulse SCLK OFF, Clear SDATA */
			writebits_w( &((board)->regs.cmd1), EEPROMCS | SCLK, SDATA);				/* Pulse SCLK ON, Clear SDATA */
		}
		else
		{
			writebits_w( &((board)->regs.cmd1), EEPROMCS | SDATA, SCLK);				/* Pulse SCLK OFF, Set SDATA */
			writebits_w( &((board)->regs.cmd1), EEPROMCS | SCLK | SDATA, 0);			/* Pulse SCLK OFF, Set SDATA */
		}
		printk("Sent bit %d:%d, ERR = %d\n", index, prom_input, (!(readport_w(&(board->regs.stat1)) & EEPROMCD)));
	}

	/*
	 * Check the status register for an error while shifting in the mode.
	 */
	if (!(readport_w(&(board->regs.stat1)) & EEPROMCD))
		rc = -1;
	writebits_w( &((board)->regs.cmd1), EEPROMCS, SCLK | SDATA);						/* Pulse SCLK OFF */
	return rc;
}

static int eeprom_shift_in_val(void)
{
	int index, value = 0;

	/*
	 * The calibration constant is shifted out of the EEPROM one bit at a time
	 * with a low-to-high transition of the SCLK bit.  The serial data bits are
	 * reflected in the PROMOUT bit of the Status Register.
	 */
	writebits_w( &((board)->regs.cmd1), EEPROMCS, SCLK);									/* Pulse SCLK OFF */
	for ( index = 0; index < 8; index++)
	{
		writebits_w( &((board)->regs.cmd1), EEPROMCS | SCLK, 0);							/* Pulse SCLK ON */
		value |= ((readport_w(&(board->regs.stat1)) & EEPROMDATA) ? 1 : 0) << (7 - index); /* Read a data bit */
		printk("Read bit %d\n", readport_w(&board->regs.stat1) & EEPROMDATA);
		writebits_w( &((board)->regs.cmd1), EEPROMCS, SCLK);								/* Pulse SCLK OFF */
	}
	printk("Read value of %x\n", value);
	writebits_w( &((board)->regs.cmd1), 0, EEPROMCS | SCLK);									/* Clear EEPROMCS - Disable chip select */
	return value;
}

static void write_caldac(int dac, int value)
{
	int bitmask, bitwise_value, index;

	writebits_w( &((board)->regs.cmd1), 0, EEPROMCS | SCLK);
	writebits_w( &((board)->regs.cmd1), 0, EEPROMCS | SCLK);

	/*
	 * The calibration constant is written to the DAC one bit at a time
	 * with a low-to-high transition of the SCLK bit.  The serial data bits are
	 * set/cleared in the SDATA bit of the Command Register 1.  The 4 most
	 * significant bits of the calibration constant indicate the DAC number.
	 */
	if (dac != DAC_IN_PREG_OFFSET)
	{
		bitmask = 0x0800;
		value &= 0x00FF;
		value |= dac << 8;
		for ( index = 0; index < 12; index++)
		{
			bitwise_value = ((value & (bitmask >> index)) ? 1 : 0);
			if (bitwise_value == 0)
			{
				writebits_w( &((board)->regs.cmd1), 0, SCLK | SDATA);					/* Pulse SCLK OFF, Set SDATA appropriately */
				writebits_w( &((board)->regs.cmd1), SCLK, SDATA);						/* Pulse SCLK ON, Leave SDATA unchanged */
			}
			else
			{
				writebits_w( &((board)->regs.cmd1), SDATA, SCLK);						/* Pulse SCLK OFF, Set SDATA appropriately */
				writebits_w( &((board)->regs.cmd1), SDATA | SCLK, 0);					/* Pulse SCLK ON, Leave SDATA unchanged */
			}
		}
	}
	else
	{
		bitmask = 0x8000;
		value &= 0x0FFF;
		value |= dac << 12;
		for ( index = 0; index < 16; index++)
		{
			bitwise_value = ((value & (bitmask >> index)) ? 1 : 0);
			if (bitwise_value == 0)
			{
				writebits_w( &((board)->regs.cmd1), 0, SCLK | SDATA);					/* Pulse SCLK OFF, Set SDATA appropriately */
				writebits_w( &((board)->regs.cmd1), SCLK, SDATA);						/* Pulse SCLK ON, Leave SDATA unchanged */
			}
			else
			{
				writebits_w( &((board)->regs.cmd1), SDATA, SCLK);						/* Pulse SCLK OFF, Set SDATA appropriately */
				writebits_w( &((board)->regs.cmd1), SDATA | SCLK, 0);					/* Pulse SCLK ON, Leave SDATA unchanged */
			}
		}
	}

	/*
	 * Pulsing the CALDAC?LD loads the specified calibration DAC with the bits clocked in by SCLK.
	 */
	writebits_w( &((board)->regs.cmd1), 0, SCLK);
	if (dac != DAC_IN_PREG_OFFSET)
		outb_p( 0x00, DAQ_CALDAC0LD);
	else
		outb_p( 0x00, DAQ_CALDAC1LD);
}

