/*
 *      $Header: atMIO16x.h,v 0.5 1995/10/24 16:32:20 amish$
 *      Revision 0.5  1996/10/9 11:33:00  amish
 *      Revision 0.5
 *
 *      Revision 0.6  1998/4/6 17:30:00  amish
 *      Revision 0.6
 *
 */
 
#ifndef _LINUX_ATMIO16X_H
#define _LINUX_ATMIO16X_H

#include <sys/ioctl.h>

/* minor device numbers (make sure to get these right when doing mknod :) */
#ifndef DAQ_ATMIO16X
#define DAQ_ATMIO16X		0	/* Card master device		*/
#endif
#ifndef DAQ_ADC
#define DAQ_ADC			1	/* Analog to Digital Converter*/
#endif
#ifndef DAQ_DAC
#define DAQ_DAC			2	/* Digital to Analog Converter*/
#endif
#ifndef DAQ_DIO
#define DAQ_DIO		 	3	/* The 8255 digital IO ports	*/
#endif
#ifndef DAQ_NUMDEVS
#define DAQ_NUMDEVS		4	/* number of devices supported 	*/
#endif

#ifdef __KERNEL__

/* The Factory default settings - override these on the command line */
#ifndef ATMIO16X_BASE
#define	ATMIO16X_BASE		0x260
#endif
#ifndef ATMIO16X_IRQ
#define ATMIO16X_IRQ		5
#endif
#ifndef ATMIO16X_DMAA
#define ATMIO16X_DMAA		6
#endif
#ifndef ATMIO16X_DMAB
#define ATMIO16X_DMAB		7
#endif

/* refer to manual for descriptions of what they do and r/w assignments */
/* Configuration and Status group	*/
#define DAQ_CMD1		board->base
#define DAQ_CMD2		(board->base + 0x02)
#define DAQ_CMD3		(board->base + 0x04)
#define DAQ_CMD4		(board->base + 0x06)
#define DAQ_STAT1		(board->base + 0x18)
#define DAQ_STAT2		(board->base + 0x1a)
/* Analog Input group			*/
#define DAQ_FIFO		(board->base + 0x00)
#define DAQ_CFGMEM		(board->base + 0x08)
/* Analog Output group			*/
#define DAQ_DAC0REG		(board->base + 0x10)
#define DAQ_DAC1REG		(board->base + 0x12)
/* Analog Event Strobe group		*/
#define DAQ_CONFIGMEMCLR	(board->base + 0x1B)
#define DAQ_CONFIGMEMLD		(board->base + 0x1B)
#define DAQ_DAQCLR		(board->base + 0x19)
#define DAQ_DAQSTRT		(board->base + 0x1D)
#define DAQ_SINGCONV		(board->base + 0x1D)
#define DAQ_ADCCAL		(board->base + 0x1F)
/* DAC Event Strobe group		*/
#define DAQ_TMRREQCLR		(board->base + 0x1F)
#define DAQ_DACUPD		(board->base + 0x18)
#define DAQ_DACCLR		(board->base + 0x1E)
/* General Event Strobe group		*/
#define DAQ_DMACHANCLR		(board->base + 0x0B)
#define DAQ_DMATCACLR		(board->base + 0x19)
#define DAQ_DMATCBCLR		(board->base + 0x09)
#define DAQ_EXTSTROBE		(board->base + 0x1E)
#define DAQ_CALDAC0LD		(board->base + 0x0A)
#define DAQ_CALDAC1LD		(board->base + 0x1A)
/* Am9513A Counter/Timer Group A	*/
#define DAQ_9513DATA		(board->base + 0x14)
#define DAQ_9513CMD		(board->base + 0x16)
#define DAQ_9513STAT		(board->base + 0x16)
/* Digital I/O group			*/
#define DAQ_DIGIN		(board->base + 0x1C)
#define DAQ_DIGOUT		(board->base + 0x1C)
/* RTSI Switch group			*/
#define DAQ_RTSISHIFT		(board->base + 0x0C)
#define DAQ_RTSISTROBE		(board->base + 0x0E)

/* How many IO ports do we use ? */
#define ATMIO16XIOSIZE		(0x20)

/* Mask for the bit fields in each register */
/* Command Register 1 - DAQ_CMD1 */
#define	EEPROMCS		0x8000	/* select calib.constants EPROM */
#define	SDATA			0x4000  /* transm.serial data to EEPROM */
#define	SCLK			0x2000  /* clocks SDATA */
#define	SCANDIV			0x1000	/* ctrl config mem sequencing */
#define	__JJ0__			0x0800	/* */
#define	INTGATE			0x0400  /* Internal Gate */
#define	RETRIG_DIS		0x0200  /* acquis. Retrigger Disable */
#define	DAQEN			0x0100	/* Acquisition Enable */
#define	SCANEN			0x0080	/* Scan Enable */
#define	SCN2			0x0040  /* acquisition Scan Mode 2 */
#define	CNT32_16		0x0020  /* select 32/16 sample count */
#define	RTSITRIG		0x0010	/* trigger A/D over RTSI? */
#define	__JJ1__			0x0008	/* */
#define	__JJ2__			0x0004  /* */
#define	__JJ3__			0x0002  /* */
#define	__JJ4__			0x0001	/* */

/* Command Register 2 - DAQ_CMD2 */
#define	A4RCV			0x8000	/* RTSI A4 Receive */
#define	A4DRV			0x4000  /* RTSI A4 Drive */
#define	A2RCV			0x2000  /* RTSI A2 Receive */
#define	A2DRV			0x1000	/* RTSI A2 Drive */
#define	BIPDAC1			0x0800	/* Bipolar DAC 1 */
#define	BIPDAC0			0x0400  /* Bipolar DAC 0 */
#define	EXTREFDAC1		0x0200  /* External Reference for DAC 1 */
#define	EXTREFDAC0		0x0100	/* External Reference for DAC 0 */
#define	EISA_DMA		0x0080	/* EISA Computer DMA Enable */
#define	__JJ5__			0x0040  /* */
#define	DMACHBB2		0x0020  /* Secondary DMA channel select */
#define	DMACHBB1		0x0010	/* */
#define	DMACHBB0		0x0008	/* */
#define	DMACHAB2		0x0004  /* Primary DMA channel select */
#define	DMACHAB1		0x0002  /* */
#define	DMACHAB0		0x0001	/* */

/* Command Register 3 - DAQ_CMD3 */
#define	ADCDSP			0x8000	/* ADC DSP Link Enable */
#define	DIOPBEN			0x4000  /* Digital I/O Port B Enable */
#define	DIOPAEN			0x2000  /* Digital I/O Port A Enable */
#define	DMATCINT		0x1000	/* DMA Terminal Count Irq Enable */
#define	DACCMPLINT		0x0800	/* DAC Complete Interrupt Enable */
#define	DAQCMPLINT		0x0400  /* DAQ Complete Interrupt Enable */
#define	IO_INT			0x0200  /* Input/Output Interrupt Enable */
#define	DMACHA			0x0100	/* DMA Channel A Enable */
#define	DMACHB			0x0080	/* DMA Channel B Enable */
#define	ADCREQ			0x0040  /* ADC Request Enable */
#define	DAC1REQ			0x0020  /* DAC 1 Request Enable */
#define	DAC0REQ			0x0010	/* DAC 0 Request Enable */
#define	DRVAIS			0x0008	/* Drive Analog Input Sense */
#define	INTCHB2			0x0004  /* Interrupt Channel select */
#define	INTCHB1			0x0002  /* */
#define	INTCHB0			0x0001	/* */

/* Command Register 4 - DAQ_CMD4 */
#define	CLKMODEB1		0x8000	/* Clock Mode select */
#define	CLKMODEB0		0x4000  /* */
#define	DAC1DSP			0x2000  /* DAC 1 DSP Link Enable */
#define	DAC0DSP			0x1000	/* DAC 0 DSP Link Enable */
#define	DACMB3			0x0800	/* DAC Mode select */
#define	DACMB2			0x0400  /* */
#define	DACMB1			0x0200  /* */
#define	DACMB0			0x0100	/* */
#define	DACGATE			0x0080	/* DAC Update Gate */
#define	DB_DIS			0x0040  /* Double Buffering Disable */
#define	CYCLICSTOP		0x0020  /* Cyclic Stop Enable */
#define	ADCFIFOREQ		0x0010	/* ADC FIFO Request */
#define	SRC3SEL			0x0008	/* Source 3 Select */
#define	GATE2SEL		0x0004  /* Gate 2 Select */
#define	FIFO_DAC		0x0002  /* FIFO or DAC Write select */
#define	EXTTRIG_DIS		0x0001	/* External Trigger Disable */

/* Status Register 1 - DAQ_STAT1 */
#define	DAQCOMP			0x8000	/* Data Acquisition Complete */
#define	DAQPROG			0x4000  /* Data Acquisition Progress */
#define	ADCFIFOHF		0x2000  /* ADC FIFO Half-Full Flag */
#define	ADCFIFOEF		0x1000	/* ADC FIFO Empty Flag */
#define	DMATCA			0x0800	/* DMA Terminal Count Chan A */
#define	DMATCB			0x0400  /* DMA Terminal Count Chan B */
#define	OVERFLOW		0x0200  /* Overflow */
#define	OVERRUN			0x0100	/* Overrun */
#define	TMRREQ			0x0080	/* Timer Request */
#define	DACCOMP			0x0040  /* DAC Sequence Complete */
#define	DACFIFOFF		0x0020  /* DAC FIFO Full Flag */
#define	DACFIFOHF		0x0010	/* DAC FIFO Half Full Flag */
#define	DACFIFOEF		0x0008	/* DAC FIFO Empty Flag */
#define	EEPROMDATA		0x0004  /* EEPROM Data */
#define	EEPROMCD		0x0002  /* EEPROM Chip Deselect */
#define	CFGMEMEF		0x0001	/* Config. Memory Empty Flag */

/* Status Register 2 - DAQ_STAT2 */
#define	__JJ6__			0x8000	/* */
#define	__JJ7__			0x4000  /* */
#define	__JJ8__			0x2000  /* */
#define	__JJ9__			0x1000	/* */
#define	__JJ10__		0x0800	/* */
#define	__JJ11__		0x0400  /* */
#define	__JJ12__		0x0200  /* */
#define	__JJ13__		0x0100	/* */
#define	__JJ14__		0x0080	/* */
#define	__JJ15__		0x0040  /* */
#define	__JJ16__		0x0020  /* */
#define	__JJ17__		0x0010	/* */
#define	__JJ18__		0x0008	/* */
#define	__JJ19__		0x0004  /* */
#define	__JJ20__		0x0002  /* */
#define	ADC_BUSY		0x0001	/* ADC is busy - inverted logic */

/* Configmem register - DAQ_CFGMEM */
#define	CHAN_SE			0x8000	/* Channel Single-Ended */
#define	CHAN_AIS		0x4000  /* Channel Analog Input Sense */
#define	CHAN_CAL		0x2000  /* Channel Calibration Enable */
#define	CHAN_BIP		0x1000	/* Channel Bipolar */
#define	__JJ21__		0x0800	/* */
#define	__JJ22__		0x0400  /* */
#define	CHANSEL3		0x0200  /* Input Channel Select */
#define	CHANSEL2		0x0100	/* */
#define	CHANSEL1		0x0080	/* */
#define	CHANSEL0		0x0040  /* */
#define	CH_GAIN2		0x0020  /* Channel Gain Select */
#define	CH_GAIN1		0x0010	/* */
#define	CH_GAIN0		0x0008	/* */
#define	CHAN_LAST		0x0004  /* Channel Last Flag */
#define	CHAN_GHOST		0x0002  /* Convert but don't put in FIFO */
#define	CHAN_DSP		0x0001	/* ADC is busy - inverted logic */

/* Struct for accessing the atMIO16x registers via the I/O ports */
typedef struct DaqPort_t {
	int	value;			/* Saved value of this register	*/
	int	address;		/* Port address of this register*/
} DaqPort_t;

typedef struct CmdRegs_t {
	DaqPort_t	cmd1;		/* atMIO16x command port		*/
	DaqPort_t	cmd2;		/* atMIO16x command port		*/
	DaqPort_t	cmd3;		/* atMIO16x command port		*/
	DaqPort_t	cmd4;		/* atMIO16x command port		*/
	DaqPort_t	stat1;		/* atMIO16x status port		*/
	DaqPort_t	stat2;		/* atMIO16x status port		*/
} CmdRegs_t;

typedef struct atMIO16x_t {
	int		major;		/* Major # for this card	*/
	int		present;	/* A atMIO16x card is present	*/
	int		open;		/* The device is open		*/
	int		stat1;		/* Device status		*/
	int		base;		/* Base address of card		*/
	int		irq;		/* Interrupt number of card	*/
	int		dma;		/* DMA channel for DMA t/fers	*/
	int		dmab;		/* DMA channel B for DMA t/fers	*/
	int		rtsi_aside;	/* RTSI switch control register contents */
	int		rtsi_bside;	/* RTSI switch control register contents */
	CmdRegs_t	regs;		/* The board command registers	*/
	struct DAC_t	*dac;	/* The Digital to Analog Conv	*/
	struct ADC_t	*adc;	/* The Analog to Digital Conv */
	struct DIO_t	*dio;	/* Digital I/O ports		*/
} atMIO16x_t;

extern atMIO16x_t atMIO16x;

/*
 * Macros to access atMIO16x control ports
 * We need to store the settings of each control register. This enables
 * us to set/clear specific bits, without affecting other control bits.
 */

/* Set the bits in the register at port address according to mask */
#define setbits_w( port, mask )	\
	{ unsigned long flags; save_flags( flags ); cli(); outw_p( (port)->value |= (mask), (port)->address ) ; SLOW_DOWN_IO ; restore_flags( flags ) ; }
#define setbits_b( port, mask )	\
	{ unsigned long flags; save_flags( flags ); cli(); outb_p( (port)->value |= (mask), (port)->address ) ; SLOW_DOWN_IO ; restore_flags( flags ) ; }

/* Clear the bits in the register at port address according to mask */
#define clearbits_w( port, mask )	\
	{ unsigned long flags; save_flags( flags ); cli(); outw_p( (port)->value &= ~(mask), (port)->address ) ; SLOW_DOWN_IO; restore_flags( flags ); }
#define clearbits_b( port, mask )	\
	{ unsigned long flags; save_flags( flags ); cli(); outb_p( (port)->value &= ~(mask), (port)->address ) ; SLOW_DOWN_IO; restore_flags( flags ); }

/* Set the bits in the register at port address according to set/clear mask */
#define writebits_w( port, setmask, clearmask )	\
	{ unsigned long flags; save_flags( flags ); cli(); outw_p( (port)->value = ((port)->value | (setmask)) & ~(clearmask),  \
	      (port)->address ) ; SLOW_DOWN_IO; restore_flags( flags ); }
#define writebits_b( port, setmask, clearmask )	\
	{ unsigned long flags; save_flags( flags ); cli(); outb_p( (port)->value = ((port)->value | (setmask)) & ~(clearmask),  \
	      (port)->address ) ; SLOW_DOWN_IO; restore_flags( flags ); }

/* Write a word to the designated port */
#define writeport_w( port, word )	\
	{ unsigned long flags; save_flags( flags ); cli(); outw_p( (port)->value = (word), (port)->address ) ; SLOW_DOWN_IO; restore_flags( flags ) ; }
#define writeport_b( port, word )	\
	{ unsigned long flags; save_flags( flags ); cli(); outb_p( (port)->value = (word), (port)->address ) ; SLOW_DOWN_IO; restore_flags( flags ) ; }

/* Read a word from the designated port */
#define readport_w( port )	\
	((port)->value = inw_p( (port)->address ))
#define readport_b( port )	\
	((port)->value = inw_p( (port)->address ))

#define BOARD( major )	(board)

#define GETDAC( minor )		\
	(board->dac)

#define GETADC( minor )		\
	(board->adc)

#define GETDIO( minor )		\
	(board->dio)

/* Global Variables */
extern atMIO16x_t *board;

extern struct file_operations counter_fops;
extern struct file_operations dac_fops;
extern struct file_operations adc_fops;
extern struct file_operations dio_fops;

#ifdef DBG_ATMIO16X
#define	dprintk( a )	printk a
#define dprintregs()		\
	dprintk(( "Registers: cmd1=0x%x cmd2=0x%x cmd3=0x%x cmd4=0x%x stat1=0x%x\n", board->regs.cmd1.value, board->regs.cmd2.value, board->regs.cmd3.value, board->regs.cmd4.value, readport_w( &(board->regs.stat1) ) ))

#else  /* DBG_ATMIO16X */
#define dprintregs()
#define dprintk( a )
#endif /* DBG_ATMIO16X */

void atmio16x_close( struct inode *inode, struct file *file );
int atmio16x_ioctl_check( unsigned int cmd, unsigned long arg );

#endif /* __KERNEL__ */

/* The magic number for ATMIO16X ioctl codes
   - see /usr/src/linux/MAGIC & /usr/src/linux/include/asm/iotcl.h */
#ifndef ATMIO16X_MAGIC
#define ATMIO16X_MAGIC		'l'
#endif

/* Ioctl values */
#define ATMIO16X_STATUS		_IO ( ATMIO16X_MAGIC, 1 )

/*
** ADC-specific structures & definitions
*/
typedef struct ADC_chan_t
{
	unsigned char channel;
	unsigned char gain;
	unsigned char inputmode;
	unsigned char ghost;
	unsigned char unipolar;
} ADC_chan_t;

#define ADC_TRIGGERMODE_PRE 0
#define ADC_TRIGGERMODE_POST 1

#define ADC_INPUTMODE_DIFF 0
#define ADC_INPUTMODE_RSE 1
#define ADC_INPUTMODE_NRSE 2

/* Allowed values for ADC_SET_POLARITY */
#define ADC_UNIPOLAR		0
#define ADC_BIPOLAR		1

/* Allowed values for ADC_SE_D */
#define ADC_SINGLEENDED		0
#define ADC_DIFFERENTIAL	1

#define ADC_NUM_CHANNELS	16

/* ADC ioctl codes */
#define ADC_CLEARCHANS		_IO ( ATMIO16X_MAGIC, 40 )
#define ADC_ADDCHAN			_IO ( ATMIO16X_MAGIC, 42 )
#define ADC_SETSR			_IO ( ATMIO16X_MAGIC, 44 )
#define ADC_GETSR			_IOR( ATMIO16X_MAGIC, 45, unsigned long )
#define ADC_STARTACQ		_IO ( ATMIO16X_MAGIC, 46 )
#define ADC_STOPACQ			_IO ( ATMIO16X_MAGIC, 48 )
#define ADC_SETCLOCK		_IO ( ATMIO16X_MAGIC, 50 )
#define ADC_GETCLOCK		_IOR( ATMIO16X_MAGIC, 51, unsigned long )
#define ADC_READFIFO		_IOR( ATMIO16X_MAGIC, 52, unsigned long )
#define ADC_SETCNT			_IO ( ATMIO16X_MAGIC, 53 )
#define ADC_SETTRIGMODE		_IO ( ATMIO16X_MAGIC, 54 )
#define ADC_USEDMA			_IO ( ATMIO16X_MAGIC, 55 )
#define ADC_INIT			_IO ( ATMIO16X_MAGIC, 59 )
#define ADC_SCONV			_IO ( ATMIO16X_MAGIC, 60 )
#define ADC_TRIGGER			_IO ( ATMIO16X_MAGIC, 61 )
#define ADC_STATUS			_IOR ( ATMIO16X_MAGIC, 62, unsigned long )
#define ADC_SETBUFSIZE		_IO ( ATMIO16X_MAGIC, 63 )
#define ADC_GETBUFSIZE		_IOR ( ATMIO16X_MAGIC, 64, unsigned long )
#define ADC_CALIBRATE		_IO ( ATMIO16X_MAGIC, 65 )
#define ADC_READEEPROM		_IOR ( ATMIO16X_MAGIC, 66, unsigned long )


/*
** DAC-specific structures & definitions
*/
#define DAC_INIT			_IO ( ATMIO16X_MAGIC, 21 )
#define DAC_STATUS			_IOR( ATMIO16X_MAGIC, 22, unsigned long )
#define DAC_SETCLOCK		_IO ( ATMIO16X_MAGIC, 23 )
#define DAC_GETCLOCK		_IOR( ATMIO16X_MAGIC, 24, unsigned long )
#define DAC_SETSR			_IO ( ATMIO16X_MAGIC, 25 )
#define DAC_GETSR			_IOR( ATMIO16X_MAGIC, 26, unsigned long )
#define DAC_START			_IO ( ATMIO16X_MAGIC, 27 )
#define DAC_STOP			_IO ( ATMIO16X_MAGIC, 28 )
#define DAC_TRIGGER			_IO ( ATMIO16X_MAGIC, 29 )
#define DAC_GETWRITETIME	_IOR( ATMIO16X_MAGIC, 30, unsigned long )
#define DAC_STRIGGER		_IO ( ATMIO16X_MAGIC, 31 )
#define DAC_GETTOTAL		_IOR( ATMIO16X_MAGIC, 32, unsigned long )
#define DAC_SETTIMEMARK		_IO ( ATMIO16X_MAGIC, 32 )
#define DAC_SETCHANS		_IO ( ATMIO16X_MAGIC, 33 )
#define DAC_GETCHANS		_IOR( ATMIO16X_MAGIC, 34, unsigned long )

/* DAC encoding mode (bipolar or unipolar output) */
#define DAC_UNIPOLAR		0
#define DAC_BIPOLAR		1

#define RTSI_A0		0
#define RTSI_A1		1
#define RTSI_A2		2
#define RTSI_A3		3
#define RTSI_A4		4
#define RTSI_A5		5
#define RTSI_A6		6

#define RTSI_B0		0
#define RTSI_B1		1
#define RTSI_B2		2
#define RTSI_B3		3
#define RTSI_B4		4
#define RTSI_B5		5
#define RTSI_B6		6

/*
** DIO-specific structures & definitions
*/

/* Ioctl values for Digital IO modes */
#define DIO_SETMODE		_IO ( ATMIO16X_MAGIC, 80 )
#define	DIO_SETBITS		_IO ( ATMIO16X_MAGIC, 81 )	
#define DIO_GETBITS		_IOR( ATMIO16X_MAGIC, 82, unsigned long )

#define DIOMODE_PORTA_INPUT 	1
#define DIOMODE_PORTA_OUTPUT	2
#define DIOMODE_PORTB_INPUT		4
#define DIOMODE_PORTB_OUTPUT	8

#endif /* _LINUX_ATMIO16X_H */

