/*
 *      $Header: atMIO16x_init.c,v 0.5 1995/10/24 16:32:20 amish$
 *      Revision 0.5  1996/10/9 11:33:00  amish
 *      Revision 0.5
 *
 *      Revision 0.6  1998/4/6 17:30:00  amish
 *      Revision 0.6
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/malloc.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <stdlib.h>

#include "adc.h"
#include "dac.h"
#include "dio.h"
#include "atMIO16x.h"

char kernel_version[] = UTS_RELEASE;

#define MAXINTERRUPTLOOP	10

#ifndef ATMIO16X_MAJOR
#define ATMIO16X_MAJOR		60
#endif

atMIO16x_t *board = NULL;

static int atmio16x_open( struct inode *inode, struct file *file )
{
	int	minor = MINOR( inode->i_rdev );
	int	stat = 0;

	if (board == NULL)
		return -(ENODEV);

	if (minor == DAQ_ADC)
	{
		file->f_op = &adc_fops;
	}
	else if (minor == DAQ_DAC)
	{
		file->f_op = &dac_fops;
	}
	else if (minor == DAQ_DIO)
	{
		file->f_op = &dio_fops;
	}
	else if (minor != DAQ_ATMIO16X)
	{
		return -(ENODEV);
	}
	stat = file->f_op->open( inode, file );
	MOD_INC_USE_COUNT;

	return stat;
}

void atmio16x_close( struct inode *inode, struct file *file )
{
	MOD_DEC_USE_COUNT;
}

int atmio16x_ioctl_check( unsigned int cmd, unsigned long arg )
{
	int	len, err;

	if (cmd & IOC_INOUT)
	{
		/* Have to validate the address given by the process. */
		len = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;

		if (cmd & IOC_IN)
		{
			if ((err = verify_area (VERIFY_READ, (void *) arg, len)) < 0)
				return err;
		}

		if (cmd & IOC_OUT)
		{
			if ((err = verify_area (VERIFY_WRITE, (void *) arg, len)) < 0)
				return err;
		}
	}

	return 0;
}

static int atmio16x_ioctl( struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg )
{
	int	stat = 0;

	if ((stat = atmio16x_ioctl_check( cmd, arg )) != 0)
		return stat;

	switch (cmd)
	{
		case ATMIO16X_STATUS:
			/* Enable/disable ADC input channel scanning */
			dprintregs();
			break;
		default:
			stat = -(ENOSYS);
			break;
	}

	return stat;
}

static struct file_operations atmio16x_fops =
{
	NULL,				/* seek */
	NULL,				/* read */
	NULL,				/* write */
	NULL,				/* readdir */
	NULL,				/* select */
	atmio16x_ioctl,		/* ioctl */
	NULL,				/* mmap */
	atmio16x_open,		/* open */
	atmio16x_close		/* release */
};

#ifdef OLDLINUX
static void daq_interrupt( int irq, struct pt_regs *registerp )
#else
static void daq_interrupt( int irq, void *dev_id, struct pt_regs *registerp )
#endif
{
	int	i,	status;

	i = 0;
	do {
		status = board->stat1 = readport_w( &(board->regs.stat1) );
/*		dprintk(( "atMIO16x: DAQ IRQ called, status = 0x%x\n", status ));  */
		if (status & (ADCFIFOEF | DAQCOMP | DMATCA | OVERFLOW | OVERRUN))
		{
			adc_interrupt( board->adc, status );
		}
		if (status & (DACCOMP | DMATCB))
		{
			dac_interrupt( board->dac, status );
		}
		if (++i > MAXINTERRUPTLOOP)
		{
			printk( "atMIO16x: ISR looped %d times: status = 0x%x\n", i, status );
			break;
		}
		status = board->stat1 = readport_w( &(board->regs.stat1) );
	} while (status & (DAQCOMP | DACCOMP | DMATCA | DMATCB | OVERFLOW | OVERRUN));
}

int atmio16x_major	= ATMIO16X_MAJOR;
int atmio16x_base 	= ATMIO16X_BASE;
int atmio16x_irq	= ATMIO16X_IRQ;
int atmio16x_dmaa	= ATMIO16X_DMAA;
int atmio16x_dmab	= ATMIO16X_DMAB;

int init_module( void )
{
	int 		i, irqno;
	CmdRegs_t 	*regs;

	printk( "atMIO16x: initializing...\n" );

	if ((board = (atMIO16x_t *)kmalloc( sizeof( atMIO16x_t ), GFP_KERNEL )) == NULL)
		return -(ENOMEM);

	board->major   = atmio16x_major;
	board->base    = atmio16x_base;
	board->irq     = atmio16x_irq;
	board->dma     = atmio16x_dmaa;
	board->dmab     = atmio16x_dmab;
	board->stat1   = 0;
	board->open    = 0;
	board->present = 1;
	board->rtsi_aside = 0;
	board->rtsi_bside = 0;

	regs = &(board->regs);

	/* Check access to the io ports */
	if (check_region( board->base, ATMIO16XIOSIZE )) {
		printk( "atMIO16x: IO ports 0x%x to 0x%x already in use\n",
		        board->base, board->base + ATMIO16XIOSIZE - 1 );
		printk( "atMIO16x: Driver unloading\n" );

		return -(EIO);
	}

	/* Load the addresses of the command ports */
	regs->cmd1.address = DAQ_CMD1;
	regs->cmd2.address = DAQ_CMD2;
	regs->cmd3.address = DAQ_CMD3;
	regs->cmd4.address = DAQ_CMD4;
	regs->stat1.address = DAQ_STAT1;
	regs->stat2.address = DAQ_STAT2;

	/*
	** Initialize the atMIO16x Card (Manual 5-2).
	** Note: we assume card is present before testing if this is so... 
	*/
	regs->stat1.value = 0;
	regs->stat2.value = 0;
	writeport_w( &(regs->cmd1), 0x00 );
	writeport_w( &(regs->cmd2), 0x00 );
	writeport_w( &(regs->cmd3), 0x00 );
	writeport_w( &(regs->cmd4), 0x00 );
	inb_p( DAQ_CONFIGMEMCLR );
	inb_p( DAQ_DAQCLR );
	outb_p( 0x00, DAQ_DMATCACLR );
	inb_p( DAQ_DMATCBCLR );
	inb_p( DAQ_DMACHANCLR );
	inb_p( DAQ_DACCLR );
	inb_p( DAQ_TMRREQCLR );
	outb_p( 0x00, DAQ_ADCCAL );	/* Initiate ADC Calibration -- See Documentation! */

	/*
	** Initialize the Am9513A Counter/Timer chip
	*/
	outw_p( 0xffff, DAQ_9513CMD );	/* Issue Master Reset operation */
	outw_p( 0xffef, DAQ_9513CMD );	/* Enable 16-bit access mode */
	outw_p( 0xff17, DAQ_9513CMD );	/* Point to the master mode reg. */
	outw_p( 0xf000, DAQ_9513DATA );	/* Load the master mode value */
	for (i=1; i <= 5; i++) {
		outw_p( 0xff00+i, DAQ_9513CMD );	/* Point to cntr mode reg. */
		outw_p( 0x0004, DAQ_9513DATA );	/* Store cntr mode value */
		outw_p( 0xff08+i, DAQ_9513CMD );	/* Point to cntr load reg. */
		outw_p( 0x0003, DAQ_9513DATA );	/* Store inactive count */
	}
	outw_p( 0xff5f, DAQ_9513CMD );	/* Load all counters */

	/*
	** Check the TMRREQ flag has been cleared by above operations
	*/
	if ((i = readport_w( &(regs->stat1) )) & TMRREQ) {
		printk( "atMIO16x: Failed to init board at base addr: 0x%x\n", board->base );
		return -(EIO);
	}

	/*
	** Tell the card the DMA channels and IRQ channels to use
	*/

	/* Set the DMA channel A */
	writebits_w( &((board)->regs.cmd2), board->dma, (~(board->dma) & (DMACHAB2|DMACHAB1|DMACHAB0)));

	/* Set the DMA channel B */
	writebits_w( &((board)->regs.cmd2), board->dmab << 3, (~(board->dmab << 3) & (DMACHBB2|DMACHBB1|DMACHBB0)));

	/* Set the IRQ channel */
	irqno = 0;
	if (board->irq == 3) irqno = 0;
	else if (board->irq == 4) irqno = 1;
	else if (board->irq == 5) irqno = 2;
	else if (board->irq == 7) irqno = 3;
	else if (board->irq == 10) irqno = 4;
	else if (board->irq == 11) irqno = 5;
	else if (board->irq == 12) irqno = 6;
	else if (board->irq == 15) irqno = 7;
	writebits_w( &((board)->regs.cmd3), irqno, (~(irqno) & (INTCHB2|INTCHB1|INTCHB0)));

	/*
	** Initialize the ADC structure
	*/
	if ((board->adc = get_adc( DAQ_ADC )) == NULL) {
		kfree( board );
		return -(ENOMEM);
	}

	/*
	** Initialize the DAC structure
	*/
	if ((board->dac = get_dac( DAQ_DAC )) == NULL) {
		kfree( board->adc );
		kfree( board );
		return -(ENOMEM);
	}

	/*
	** Initialize the DIO structure
	*/
	if ((board->dio = get_dio( DAQ_DIO )) == NULL) {
		kfree( board->adc );
		kfree( board->dac );
		kfree( board );
		return -(ENOMEM);
	}

	/*
	** Reserve the ports for our use
	*/
	request_region( board->base, ATMIO16XIOSIZE, "atMIO16x" );

	if (register_chrdev( board->major, "atMIO16x", &atmio16x_fops )) {
		printk( "atMIO16x: Can't get major %d\n", board->major );
		return -(EINVAL);
	} else {
		printk( "atMIO16x Card detected. Installed driver V0.1\n" );
	}

	if (request_irq( board->irq, daq_interrupt, SA_INTERRUPT, "atMIO16x", NULL )) {
		printk( "atMIO16x: can't request IRQ %d\n", board->irq );
		return -(EIO);
	}

	if (request_dma( board->dma, "atMIO16x" )) {
		printk( "atMIO16x: can't request DMA chan %d\n", board->dma );
		return -(EIO);
	}
	if (request_dma( board->dmab, "atMIO16x" )) {
		printk( "atMIO16x: can't request DMA chan %d\n", board->dmab );
		return -(EIO);
	}

	printk( "atMIO16x: initialized board!\n" );
	return 0;
}

void cleanup_module(void)
{
	if (MOD_IN_USE) {
		printk("atMIO16x: device busy, remove delayed\n");
	}

	free_adc( board->adc );
	dprintk(( "atMIO16x: ADC released\n" ));

	free_dac( board->dac );
	dprintk(( "atMIO16x: DAC released\n" ));

	free_dio( board->dio );
	dprintk(( "atMIO16x: DIO released\n" ));

	free_irq( board->irq, NULL );
	dprintk(( "atMIO16x: irq released\n" ));

	free_dma( board->dma );
	dprintk(( "atMIO16x: dma released\n" ));

	free_dma( board->dmab );
	dprintk(( "atMIO16x: dmab released\n" ));

	release_region( board->base, ATMIO16XIOSIZE );
	dprintk(( "atMIO16x: ioports released\n" ));

	if (unregister_chrdev( board->major, "atMIO16x" ) != 0) {
		printk( "atMIO16x: device busy or other module error \n" );
	} else {
		printk( "atMIO16x: succesfully removed \n" );
	} 
	dprintk(( "atMIO16x: device major released\n" ));

	kfree( board );
	dprintk(( "atMIO16x: Board released\n" ));
}

