/*
 *      $Header: dac.c,v 0.5 1995/10/24 16:32:20 amish$
 *      Revision 0.5  1996/10/9 11:33:00  amish
 *      Revision 0.5
 *
 *      Revision 0.6  1998/4/6 17:30:00  amish
 *      Revision 0.6
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/dma.h>

#include "dac.h"
#include "atMIO16x.h"

#define INCREMENT64(num) {if ((num)[0] == ULONG_MAX) {(num)[0] = 0; (num)[1]++;} else {(num)[0]++;}}
#define ADD64(num, arg) {if ((num)[0] > ULONG_MAX - arg) {(num)[0] = arg-1; (num)[1]++;} else {(num)[0]+=arg;}}

#define MAX_DMA_SIZE	(128 * 1024)

#ifndef DACBUFSIZE
#define DACBUFSIZE	8192
#endif

static struct semaphore mutex = MUTEX;

int dacbufsize = DACBUFSIZE;

/* Accessing Board command and status registers */
#define DACstatus( board )		\
	(readport_w( &((board)->regs.stat1) ))

/* Clear the DAC - strobe clear register */
#define ClearDAC( dac )			\
	{ inb_p( DAQ_DACCLR ); inb_p( DAQ_TMRREQCLR ); }

/* Clear a DAC DMA Terminal Count interrupt */
#define ClearDMAInterrupt( dac )	\
	inb_p( DAQ_DMATCBCLR )

/* Disable all DAC interrupt generation */
#define DisableDACInterrupts( dac )		\
	clearbits_w( &((board)->regs.cmd3), DACCMPLINT | DMACHB )

/* Disable DAC conversions */
#define DisableDACConversions( dac )		\
	setbits_w( &((board)->regs.cmd4), DACGATE )

/* Enable DAC conversions */
#define EnableDACConversions( dac )		\
	clearbits_w( &((board)->regs.cmd4), DACGATE )


#define ZeroDAC( dac ) \
{\
	outw_p(0x0000, DAQ_DAC0REG);	\
	outw_p(0x0000, DAQ_DAC1REG);	\
	outw_p(0x0000, DAQ_DACUPD);	\
}

#define Reset9513Counter(ctr) \
{\
	outw_p(0xffc0 + (1<<((ctr)-1)), DAQ_9513CMD);	\
	outw_p(0xff00 + (ctr), DAQ_9513CMD);	\
	outw_p(0x0004, DAQ_9513DATA);	\
	outw_p(0xff08 + (ctr), DAQ_9513CMD);	\
	outw_p(0x0003, DAQ_9513DATA);	\
	outw_p(0xff40 + (1<<((ctr)-1)), DAQ_9513CMD);	\
	outw_p(0xff40 + (1<<((ctr)-1)), DAQ_9513CMD);	\
}

#define STATIC

STATIC int allocate_dac_buffers( DAC_t *dac )
{
	int i;
	short *ptr;

	if (dac->bufsize == 0)
	{
		dac->frontbuf = NULL;
		dac->backbuf = NULL;
		dac->zerobuf = NULL;
		return 0;
	}

	/* Allocate buffers for DMA transfers */
	if (dac->bufsize * sizeof( short ) > MAX_DMA_SIZE)
	{
		printk( "atMIO16x: Requested DMA buffer size, %d, too large\n", dac->bufsize );
		dac->bufsize = MAX_DMA_SIZE / sizeof( short );
		printk( "atMIO16x: Using DMA buffer size, %d\n", dac->bufsize );
	}
	dac->frontbuf = kmalloc( dac->bufsize * sizeof( short ), GFP_KERNEL | GFP_DMA );
	if (dac->frontbuf == NULL)
	{
		printk( "DAC: Unable to allocate DMA buffer (%d)\n", dac->bufsize * sizeof(short) );
		dac->bufsize = 0;
		return -(ENOMEM);
	}
	dac->backbuf = kmalloc( dac->bufsize * sizeof( short ), GFP_KERNEL | GFP_DMA );
	if (dac->backbuf == NULL)
	{
		kfree( dac->frontbuf );
		dac->bufsize = 0;
		printk( "DAC: Unable to allocate DMA buffer (%d)\n", dac->bufsize * sizeof(short) );
		return -(ENOMEM);
	}
	dac->zerobuf = kmalloc( dac->bufsize * sizeof( short ), GFP_KERNEL | GFP_DMA );
	if (dac->zerobuf == NULL)
	{
		kfree( dac->frontbuf );
		kfree( dac->backbuf );
		dac->bufsize = 0;
		printk( "DAC: Unable to allocate DMA buffer (%d)\n", dac->bufsize * sizeof(short) );
		return -(ENOMEM);
	}
	for (i=0, ptr = dac->zerobuf; i < dac->bufsize; i++)
		ptr[i] = 0;
	return 0;
}

STATIC int select_dac(struct inode *inode, struct file *file, int sel_type, select_table *wait)
{
	DAC_t *dac = BOARD(MAJOR(inode->i_rdev))->dac;
	int	retval;

	/*
	** We are always ready (ie: never wait) for input or exceptions.
	** If we can immediately write a buffer (bufcount == 0), then we
	** don't wait on output either.  In all other cases, we wait.
	** This means we call select_wait() (which returns immediately), and return 0,
	** so that our calling function (do_select()) can wait.
	*/
	if (sel_type != SEL_OUT)
		return 1;
	retval = 0;
	cli();
	if (dac->bufcount == 0)
		retval = 1;
	else
		select_wait( &(dac->wait), wait );
	sti();
	return retval;
}

STATIC int init_dac( DAC_t *dac )
{
	/* Initialise the DAC registers */
#ifdef USECOUNTER2
	Reset9513Counter(2);						/* Reset counter 2 */
#else
	Reset9513Counter(5);						/* Reset counter 5 */
#endif
	DisableDACInterrupts(dac);
	DisableDACConversions(dac);
	ClearDAC(dac);
	ZeroDAC(dac);
	return 0;
}

/* NOTE: Can be called with interrupts disabled - don't do anything to re-enable them! */
int trigger_dac( DAC_t *dac )
{
#ifdef USECOUNTER2
	outw_p( 0xff22, DAQ_9513CMD );				/* Arm Counter 2 */
#else
	outw_p( 0xff30, DAQ_9513CMD );				/* Arm Counter 5 */
#endif
	EnableDACConversions(dac);
	return 0;
}

static int write_dac(struct inode *inode, struct file *file, const char *buffer, int count)
{
	DAC_t *dac;
	int	donecount, status, retval, chunksize;
	unsigned long flags;

/*	dprintk(( "atMIO16x+: write_dac()\n" )); */
	if ((dac = GETDAC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	/* Lockout other processes from accessing the DAC */
	down( &(dac->sem) );

	/*
	** count is the # of samples to write to the DAC.
	** we loop, sending chunksize samples at a time
	** we are done when everything has been sent, or
	** there was an error.
	** In the O_NONBLOCK case, we see if we can send
	** a single chunk.  If not, then return w/o waiting.
	** If so, send it, and return the number of BYTES
	** actually sent.
	** If NOT O_NONBLOCK, then block until we have 
	** sent everything we were asked to.
	*/
	save_flags( flags );
	cli();

	count /= sizeof( short );
	donecount = 0;
	retval = 0;

	while (donecount < count)
	{
		if (dac->bufcount != 0)
		{
			if (file->f_flags & O_NONBLOCK)
			{
				retval = -(EAGAIN);
				break;
			}
			interruptible_sleep_on( &(dac->wait) );
			status = DACstatus( board );
			if (current->signal & ~current->blocked)			/* Interrupted by user */
			{
				ClearDAC(dac);
				dprintk(( "DAC: interrupted\n" ));
				retval = -(ERESTARTSYS);
				break;
			}
		}

		if (dac->bufcount == 0)
		{
			chunksize = (dac->bufsize > (count - donecount)) ? (count - donecount) : dac->bufsize;
			memcpy_fromfs(dac->backbuf, buffer + (sizeof(short) * donecount), chunksize * sizeof(short));
			donecount += chunksize;
			dac->bufcount = chunksize;
			if (dac->timemarkflag == 1)
			{
				dac->writetime = dac->total + 1;
				dac->timemarkflag = 0;
			}
			if (file->f_flags & O_NONBLOCK)
				break;
		}
	}

	restore_flags( flags );
	up( &(dac->sem) );
	return (retval == 0) ? (donecount * sizeof(short)) : (retval);
}

STATIC int start_dac( DAC_t *dac )
{
	int interval;
	int clockspeed = dac->clockspeed;
#ifdef USECOUNTER2
	int i, mask, bit;
#endif
	int arg;

	dprintk(( "atMIO16x: start_dac() called\n" ));
	dprintregs();

	/*
	** Before doing anything else,
	** validate the desired # of chans and samplerate
	** with the actual capability of the board
	*/
	interval = clockspeed / dac->samplerate;
	if (interval * dac->samplerate != clockspeed)
		return -(EINVAL);
	if (dac->chans == 3)
	{
		if (interval % 2 == 1)
		{
			printk("atMIO16x: start_dac(): INVALID RATE!\n");
			return -(EINVAL);
		}
		interval = interval / 2;
	}
	dprintk(( "atMIO16x: start_dac(): rate = %d, interval = %d\n", dac->samplerate, interval ));

	/*
	** Clear the Analog Output Circuitry
	*/
	setbits_w( &((board)->regs.cmd4), DACGATE );
	inb_p(DAQ_TMRREQCLR);
/*	inb_p(DAQ_DACCLR); */
	inb_p(DAQ_DMATCBCLR);
	dprintregs();

#ifdef USECOUNTER2
	/*
	** Reset the Am9513A counter 2 (which is the only one we'll use for DAC)
	*/
	Reset9513Counter(2);

	/*
	** Route the board signal OUT2 to drive the RTSI switch pin A2
	** Also, route the RTSI switch pin A4 to drive the board signal TMRTRIG*
	*/
	setbits_w( &((board)->regs.cmd2), A2DRV | A4RCV );

	/*
	** Program the RTSI switch
	** The signal on RTSI-pin A2 (OUT2) should drive RTSI-pin B6 (TRIGGER6)
	** and the signal on RTI-pin B6 (TRIGGER6) should drive RTSI-pin A4 (TMRTRIG*)
	*/
	printk("before aside = %x\n", board->rtsi_aside);
	printk("before bside = %x\n", board->rtsi_bside);
	board->rtsi_bside &= (~ (0xf << (RTSI_B6 << 2)) );
	board->rtsi_bside |= ( ((RTSI_A2 << 1) | 1) << (RTSI_B6 << 2) );
	board->rtsi_aside &= (~ (0xf << (RTSI_A4 << 2)) );
	board->rtsi_aside |= ( ((RTSI_B6 << 1) | 1) << (RTSI_A4 << 2) );
	printk("after aside = %x\n", board->rtsi_aside);
	printk("after bside = %x\n", board->rtsi_bside);

	/*
	** Program the RTSI control register
	*/
	for (i=0, mask = 1L; i < 28; i++, mask <<= 1)
	{
		bit = (board->rtsi_bside & mask) ? 1 : 0;
		outb(DAQ_RTSISHIFT, bit);
	}
	for (i=0, mask = 1L; i < 28; i++, mask <<= 1)
	{
		bit = (board->rtsi_aside & mask) ? 1 : 0;
		outb(DAQ_RTSISHIFT, bit);
	}
	outb(DAQ_RTSISTROBE, 0);

	/*
	** Program the update-interval counter (Counter 2 on the Am9513A)
	*/
	outw_p( 0xff02, DAQ_9513CMD );		/* Select Counter 2 mode reg. */
	if (clockspeed == 5000000)
		outw_p( 0x0225, DAQ_9513DATA );	/* Select 5Mhz clock */
	else if (clockspeed == 1000000)
		outw_p( 0x0b25, DAQ_9513DATA );	/* Select 1Mhz clock */
	else if (clockspeed == 100000)
		outw_p( 0x0c25, DAQ_9513DATA );	/* Select 100kHz clock */
	else if (clockspeed == 10000)
		outw_p( 0x0d25, DAQ_9513DATA );	/* Select 10kHz clock */
	else if (clockspeed == 1000)
		outw_p( 0x0e25, DAQ_9513DATA );	/* Select 1kHz clock */
	else if (clockspeed == 100)
		outw_p( 0x0f25, DAQ_9513DATA );	/* Select 100Hz clock */
	else
	{
		dprintk(( "atMIO16x: INVALID CLOCK SELECTION-using 100kHz!!\n" ));
		outw_p( 0x0c25, DAQ_9513DATA );	/* Select 100kHz clock */
	}
	outw_p( 0xff0a, DAQ_9513CMD );		/* Select Counter 2 load reg. */
	outw_p( interval, DAQ_9513DATA );	/* Store Counter 2 load value */
	outw_p( 0xff42, DAQ_9513CMD );		/* Load Counter 2 */
/*	outw_p( 0xfff2, DAQ_9513CMD );*/		/* Decrement Counter 2 */
#ifdef notdef
	outw_p( 0xff22, DAQ_9513CMD );		/* Arm Counter 2 */
#endif
#else
	/*
	** Reset the Am9513A counter 5 (which is the only one we'll use for DAC)
	*/
	Reset9513Counter(5);

	setbits_w( &((board)->regs.cmd2), A4RCV | A4DRV );

	/*
	** Program the update-interval counter (Counter 5 on the Am9513A)
	*/
	outw_p( 0xff05, DAQ_9513CMD );		/* Select Counter 5 mode reg. */
	if (clockspeed == 5000000)
		outw_p( 0x0225, DAQ_9513DATA );	/* Select 5Mhz clock */
	else if (clockspeed == 1000000)
		outw_p( 0x0b25, DAQ_9513DATA );	/* Select 1Mhz clock */
	else if (clockspeed == 100000)
		outw_p( 0x0c25, DAQ_9513DATA );	/* Select 100kHz clock */
	else if (clockspeed == 10000)
		outw_p( 0x0d25, DAQ_9513DATA );	/* Select 10kHz clock */
	else if (clockspeed == 1000)
		outw_p( 0x0e25, DAQ_9513DATA );	/* Select 1kHz clock */
	else if (clockspeed == 100)
		outw_p( 0x0f25, DAQ_9513DATA );	/* Select 100Hz clock */
	else
	{
		dprintk(( "atMIO16x: INVALID CLOCK SELECTION-using 100kHz!!\n" ));
		outw_p( 0x0c25, DAQ_9513DATA );	/* Select 100kHz clock */
	}
	outw_p( 0xff0d, DAQ_9513CMD );		/* Select Counter 5 load reg. */
	outw_p( interval, DAQ_9513DATA );	/* Store Counter 5 load value */
	outw_p( 0xff50, DAQ_9513CMD );		/* Load Counter 5 */
	outw_p( 0xfff5, DAQ_9513CMD );		/* Decrement Counter 5 */
#ifdef notdef
	outw_p( 0xff30, DAQ_9513CMD );		/* Arm Counter 5 */
#endif
#endif

	/*
	** Configure DAC 0 and DAC 1 for Bipolar operation
	** Use internal reference (+10V)
	** Configure DAC Mode for DMA access through DAC FIFO (with single requesting)
	** Enable double-buffering and de-glitching of DACs
	*/
	setbits_w( &((board)->regs.cmd2), BIPDAC1 | BIPDAC0 );
	clearbits_w( &(board->regs.cmd2), EXTREFDAC1 | EXTREFDAC0);
	setbits_w( &((board)->regs.cmd4), DACMB0 | FIFO_DAC);
	clearbits_w( &(board->regs.cmd4), DACMB1 | DACMB2 | DB_DIS );

	/* Program the DMA controller */
	cli();
	disable_dma( board->dmab );
	dac->count = dac->bufsize;
	clear_dma_ff( board->dmab );
	set_dma_mode( board->dmab, DMA_MODE_WRITE );
	set_dma_addr( board->dmab, (unsigned long)dac->zerobuf );
	set_dma_count( board->dmab, dac->count * sizeof(short) );
	enable_dma( board->dmab );
	sti();

	/*
	** At this point, the board is waiting for
	** DAC Conversions to be enabled (DACGATE),
	** and for the Counter to be armed...
	*/
	dac->writetime = 0LL;
	dac->timemarkflag = 0;
	dac->total = dac->count;

	/* Enable DMA IRQs on chan B */
	ClearDMAInterrupt(dac);
	dprintk(( "atMIO16x: start_dac() enabling IRQs\n" ));
	dprintregs();

	arg = 0;
	if (dac->chans & 1) arg |= DAC0REQ;
	if (dac->chans & 2) arg |= DAC1REQ;
	if (arg == 0) arg |= DAC0REQ;
	clearbits_w(&((board)->regs.cmd3), (DAC0REQ | DAC1REQ));
	setbits_w(&((board)->regs.cmd3), (DACCMPLINT | DMACHB | DMATCINT | arg));
	printk("atMIO16x: arg = %x\n", arg);

	dprintk(( "atMIO16x: start_dac() finished\n" ));
	dprintregs();
	return 0;
}

STATIC int ioctl_dac(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	DAC_t *dac;
	int	stat = 0;
	unsigned long long tmp64;

	if ((stat = atmio16x_ioctl_check( cmd, arg )) != 0)
		return stat;

	/* Which DAC? */
	if ((dac = GETDAC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	down( &(dac->sem) );
	switch (cmd)
	{
		case DAC_INIT:
			init_dac(dac);
			dprintk(( "atMIO16x: DAC: handled INIT ioctl\n" ));
			dprintregs();
			break;
		case DAC_START:
			start_dac( dac );
			dprintk(( "atMIO16x: DAC: handled START ioctl\n" ));
			break;
		case DAC_TRIGGER:
			trigger_dac(dac);
			dprintk(( "atMIO16x: DAC: handled TRIGGER ioctl\n" ));
			break;
		case DAC_STRIGGER:
			trigger_dac(dac);
			inb_p( DAQ_DAQSTRT );	/* trigger ADC conversions */
			dprintk(( "atMIO16x: DAC: handled STRIGGER ioctl\n" ));
			break;
		case DAC_STATUS:
			dprintregs();
			dprintk(( "watchdog = %d\n", dac->watchdog ));
			put_fs_long((unsigned long)(DACstatus(board)), (unsigned long *)arg);
			break;
		case DAC_SETCHANS:
			dac->chans = get_fs_long((unsigned long *)arg);
			if ((dac->chans & (!(1|2))) || (dac->chans == 0))
			{
				dac->chans = 1;
				stat = -(EINVAL);
			}
			printk( "chans = %d\n", dac->chans);
			break;
		case DAC_GETCHANS:
			put_fs_long( dac->chans, (unsigned long *)arg );
			break;
		case DAC_SETCLOCK:
			dac->clockspeed = get_fs_long((unsigned long *)arg);
			break;
		case DAC_GETCLOCK:
			put_fs_long( dac->clockspeed, (unsigned long *)arg );
			break;
		case DAC_SETTIMEMARK:
			dac->writetime = 0LL;
			dac->timemarkflag = 1;
			break;
		case DAC_SETSR:
			dac->samplerate = get_fs_long((unsigned long *)arg);
			break;
		case DAC_GETSR:
			put_fs_long( dac->samplerate, (unsigned long *)arg );
			break;
		case DAC_GETTOTAL:
			tmp64 = dac->total;
			if (dac->chans == 3) tmp64 >>= 1;		/* Divide by two */
			memcpy_tofs((unsigned long *)arg, &(tmp64), sizeof(unsigned long long));
			break;
		case DAC_GETWRITETIME:
			if (dac->timemarkflag == 1)
				stat = -(EINVAL);
			else
			{
				tmp64 = dac->writetime;
				if (dac->chans == 3) tmp64 >>= 1;
				memcpy_tofs((unsigned long *)arg, &(tmp64), sizeof(unsigned long long));
			}
			break;
		default:
			stat = -(ENOSYS);
			break;
	}
	up( &(dac->sem) );
	return stat;
}

STATIC int open_dac( struct inode *inode, struct file *file )
{
	DAC_t *dac;

	dprintk(( "atMIO16x: open_dac() called\n" ));
	if ((dac = GETDAC( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	/* This is an exclusive device */
	if (dac->open)
		return -(EBUSY);

	dac->open++;
	return 0;
}

STATIC void release_dac(struct inode *inode, struct file *file)
{
	DAC_t *dac;

	dprintk(( "atMIO16x: release_dac() called\n" ));
	if ((dac = GETDAC( MINOR(inode->i_rdev) )) == NULL)
		return;

	/* Shutdown the DAC on close() */

	dac->open--;
	atmio16x_close( inode, file );
}

void dac_interrupt( DAC_t *dac, int status )
{
	short *buf;

/*	printk("atMIO16x: DAC IRQ: 0x%x\n", status); */
	dac->watchdog++;
	if (status & DMATCB)
	{
		if (dac->bufcount == 0)
		{
			/* If there is no data to be played, write zeros to the DAC */
			buf = dac->zerobuf;
			dac->count = dac->bufsize;
		}
		else
		{
			/* If there is data, send it to the DAC, make other buffer available */
			buf = dac->backbuf;
			dac->backbuf = dac->frontbuf;
			dac->frontbuf = buf;
			dac->count = dac->bufcount;
		}

		if (get_dma_residue(board->dmab) != 0)
			printk("atMIO16x: DAC IRQ: residue = %d\n", get_dma_residue(board->dmab));

		/*
		** See comments in adc.c regarding FAST interrupts, and not
		** needing cli()/sti() protection here.
		*/
		set_dma_addr(board->dmab, (unsigned long)buf);
		set_dma_count(board->dmab, dac->count * sizeof(short));
		enable_dma(board->dmab);
		dac->total += dac->count;
		dac->bufcount = 0;
		ClearDMAInterrupt(dac);
		wake_up_interruptible( &(dac->wait) );
	}
	if (status & DACCOMP)
	{
		if (~(status & DACFIFOEF))
		{
			printk("atMIO16x: DAC UNDERFLOW!\n");
			dprintregs();
		}
		disable_dma(board->dmab);
		ClearDMAInterrupt(dac);

#ifdef USECOUNTER2
		Reset9513Counter(2);
#else
		Reset9513Counter(5);
#endif
		DisableDACInterrupts(dac);
		DisableDACConversions(dac);
		ClearDAC(dac);
		ZeroDAC(dac);
		wake_up_interruptible(&(dac->wait));
	} 
	return;
}

struct file_operations dac_fops =
{
	NULL,				/* seek */
	NULL,				/* read */
	write_dac,
	NULL,				/* readdir */
	select_dac,
	ioctl_dac,
	NULL,				/* mmap */
	open_dac,
	release_dac,
};

/* Functions to load and unload the DAC structure */
DAC_t *get_dac(int minor)
{
	DAC_t	*dac;

	dac = (DAC_t *)kmalloc(sizeof(DAC_t), GFP_KERNEL);
	dac->open = 0;
	dac->sem = mutex;
	dac->wait = NULL;
	dac->watchdog = 0;
	dac->bufcount = 0;
	dac->bufsize = dacbufsize;
	dac->total = 0LL;
	dac->writetime = 0LL;
	dac->timemarkflag = 0;
	dac->chans = 1;

	init_dac(dac);
	if (allocate_dac_buffers(dac) < 0)
	{
		kfree(dac);
		return NULL;
	}
	dac->count = 0;

	return dac;
}

void free_dac( DAC_t *dac )
{
	if (dac != NULL)
	{
		if (dac->frontbuf != NULL)
		{
			kfree(dac->frontbuf);
			dac->frontbuf = NULL;
		}
		if (dac->backbuf != NULL)
		{
			kfree(dac->backbuf);
			dac->backbuf = NULL;
		}
		if (dac->zerobuf != NULL)
		{
			kfree(dac->zerobuf);
			dac->zerobuf = NULL;
		}
		init_dac(dac);
		kfree(dac);
	}
}

