/*
 *      $Header: dio.c,v 0.5 1995/10/24 16:32:20 amish$
 *      Revision 0.5  1996/10/9 11:33:00  amish
 *      Revision 0.5
 *
 *      Revision 0.6  1998/4/6 17:30:00  amish
 *      Revision 0.6
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>

#include "dio.h"
#include "atMIO16x.h"

static struct semaphore mutex = MUTEX;

/* SetPortAOutput */
#define SetPortAOutput( dio )			\
	setbits_w( &((board)->regs.cmd3), DIOPAEN ) 		/* Digital I/O Port A Enable */

/* SetPortBOutput */
#define SetPortBOutput( dio )			\
	setbits_w( &((board)->regs.cmd3), DIOPBEN ) 		/* Digital I/O Port B Enable */

/* SetPortAInput */
#define SetPortAInput( dio )			\
	clearbits_w( &((board)->regs.cmd3), DIOPAEN ) 		/* Digital I/O Port A Enable */

/* SetPortBInput */
#define SetPortBInput( dio )			\
	clearbits_w( &((board)->regs.cmd3), DIOPBEN ) 		/* Digital I/O Port B Enable */

#define STATIC

STATIC int init_dio( DIO_t *dio )
{
	/* Initialise the DIO registers */
	SetPortAInput(dio);
	SetPortBInput(dio);
	dio->iomode = DIOMODE_PORTA_INPUT | DIOMODE_PORTB_INPUT;

	/* Access control variables */
	dio->sem = mutex;
	dio->wait = NULL;

	return 0;
}

STATIC int ioctl_dio(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	DIO_t 	*dio;
	int	stat = 0;
	short bits;

	if ((stat = atmio16x_ioctl_check( cmd, arg )) != 0)
		return stat;

	/* Which DIO? */
	if ((dio = GETDIO( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	down( &(dio->sem) );
	switch (cmd)
	{
		case DIO_SETMODE:
			dio->iomode = get_fs_long((unsigned long *)arg);
			if (((dio->iomode & DIOMODE_PORTA_INPUT) && (dio->iomode & DIOMODE_PORTA_OUTPUT)) || 
				((dio->iomode & DIOMODE_PORTB_INPUT) && (dio->iomode & DIOMODE_PORTB_OUTPUT)))
				stat = -(EINVAL);
			else
			{
				if (dio->iomode & DIOMODE_PORTA_INPUT)
					SetPortAInput(dio);
				if (dio->iomode & DIOMODE_PORTA_OUTPUT)
					SetPortAOutput(dio);
				if (dio->iomode & DIOMODE_PORTB_INPUT)
					SetPortBInput(dio);
				if (dio->iomode & DIOMODE_PORTB_OUTPUT)
					SetPortBOutput(dio);
			}
			break;
		case DIO_SETBITS:
			bits = get_fs_long((unsigned long *)arg);
			outw_p(bits, DAQ_DIGOUT);
			break;
		case DIO_GETBITS:
			bits = inw_p(DAQ_DIGIN);
			put_fs_long(bits, (unsigned long *)arg);
			break;
		default:
			stat = -(ENOSYS);
			break;
	}
	up( &(dio->sem) );

	return stat;
}

STATIC int open_dio( struct inode *inode, struct file *file )
{
	DIO_t 	*dio;

	dprintk(( "atMIO16x: open_dio() called\n" ));

	/* Which DIO? */
	if ((dio = GETDIO( MINOR(inode->i_rdev) )) == NULL)
		return -(EINVAL);

	dio->open++;

	return 0;
}

STATIC void release_dio(struct inode *inode, struct file *file)
{
	DIO_t 	*dio;
	int		minor;

	dprintk(( "atMIO16x: release_dio() called\n" ));

	minor = MINOR(inode->i_rdev);
	if ((dio = GETDIO( minor )) == NULL)
		return;

	dio->open--;
	if (dio->open == 0)
	{
		/* Last close - clean up */
		SetPortAInput(dio);
		SetPortBInput(dio);
	}
	atmio16x_close( inode, file );
}


struct file_operations dio_fops =
{
	NULL,			/* seek */
	NULL,			/* read */
	NULL,			/* write */
	NULL,			/* readdir */
	NULL,			/* select */
	ioctl_dio,		/* ioctl */
	NULL,			/* mmap */
	open_dio,		/* open */
	release_dio,	/* release */
};

/* Functions to load and unload the DIO structure */

DIO_t *get_dio( int minor )
{
	DIO_t *dio;

	dio = (DIO_t *)kmalloc(sizeof(DIO_t), GFP_KERNEL);
	dio->open = 0;
	init_dio(dio);

	return dio;
}

void free_dio( DIO_t *dio )
{
	if (dio != NULL)
	{
		init_dio(dio);
		kfree(dio);
	}
}

