#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "atMIO16x.h"

#define HANDERRNO(errno) {if (errno) perror(NULL);}

main(int ac, char *av[])
{
	FILE *fp;
	char buf[80];
	int i, status, fd, samplerate, clockspeed, sampcnt, cnt, tmp, total;
	ADC_chan_t tmpchan;
	short databuf[131072];

	if (ac == 1)
	{
		fprintf(stderr, "Usage: do <status> <trigger> <sconv> <read file.pcm> <clear> <add n>\n");
		fprintf(stderr, "          <setsr n> <setclock n> <setcnt n> <init> <start> <eeprom n>\n");
		fprintf(stderr, "          <sread> <pretrig> <posttrig> <useirq> <usedma> <calibrate>\n");
		return -1;
	}

	fd = open("/dev/atmio16x/adc", O_RDWR);
	if (errno != 0) perror(NULL);

	for (i=1; i < ac; i++)
	{
		if (!strcmp(av[i], "status"))
		{
			ioctl(fd, ADC_STATUS, &status);
			HANDERRNO(errno);
			printf("status = 0x%x\n", status);
		}
		else if (!strcmp(av[i], "trigger"))
		{
			ioctl(fd, ADC_TRIGGER, NULL);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "useirq"))
		{
			tmp = 0;
			ioctl(fd, ADC_USEDMA, &tmp);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "usedma"))
		{
			tmp = 1;
			ioctl(fd, ADC_USEDMA, &tmp);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "calibrate"))
		{
			ioctl(fd, ADC_CALIBRATE, NULL);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "sconv"))
		{
			ioctl(fd, ADC_SCONV, NULL);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "setclock"))
		{
			sscanf(av[++i], "%d", &clockspeed);
			ioctl(fd, ADC_SETCLOCK, &clockspeed);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "pretrig"))
		{
			tmp = ADC_TRIGGERMODE_PRE;
			ioctl(fd, ADC_SETTRIGMODE, &tmp);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "posttrig"))
		{
			tmp = ADC_TRIGGERMODE_POST;
			ioctl(fd, ADC_SETTRIGMODE, &tmp);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "setcnt"))
		{
			sscanf(av[++i], "%d", &sampcnt);
			ioctl(fd, ADC_SETCNT, &sampcnt);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "setsr"))
		{
			sscanf(av[++i], "%d", &samplerate);
			ioctl(fd, ADC_SETSR, &samplerate);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "read"))
		{
			fp = fopen(av[++i], "w");
			total = 0;
			while (((cnt = read(fd, databuf, 131072)) > 0) && (cnt != -1))
			{
				fwrite(databuf, cnt, 1, fp);
				printf("cnt = %d\n", cnt/2);
				total += cnt/2;
			}
			printf("total = %d\n", total);
			fclose(fp);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "sread"))
		{
			ioctl(fd, ADC_READFIFO, &status);
			HANDERRNO(errno);
			printf("fifoval = 0x%x\n", status);
		}
		else if (!strcmp(av[i], "clear"))
		{
			ioctl(fd, ADC_CLEARCHANS, NULL);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "init"))
		{
			ioctl(fd, ADC_INIT, NULL);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "eeprom"))
		{
			sscanf(av[++i], "%d", &tmp);
			if (ioctl(fd, ADC_READEEPROM, &tmp) != -1)
				printf("eeprom = 0x%x (%d)\n", tmp, tmp);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "add"))
		{
			sscanf(av[++i], "%d", &tmp);
			tmpchan.channel = tmp;
			tmpchan.gain = 1;
			tmpchan.inputmode = ADC_INPUTMODE_NRSE;
			tmpchan.ghost = 0;
			tmpchan.unipolar = 0;
			ioctl(fd, ADC_ADDCHAN, &tmpchan);
			HANDERRNO(errno);
		}
		else if (!strcmp(av[i], "start"))
		{
			ioctl(fd, ADC_STARTACQ, NULL);
			HANDERRNO(errno);
		}
	}
	close(fd);
}

